#!/usr/bin/env python

# flake8: noqa

from __future__ import print_function
from __future__ import absolute_import
import os
import json
import yaml
import json
import difflib

from configparser import ConfigParser

from nose.tools import *

from linchpin.context import LinchpinContext
from linchpin import LinchpinAPI
from linchpin.InventoryFilters import GenericInventory
from linchpin.tests.mockdata.contextdata import ContextData

def setup_generic_inventory_filter():
    global filter
    global res_output

    cd = ContextData()
    cd.load_config_data()
    config_path = cd.get_temp_filename()
    config_data = cd.cfg_data
    cd.write_config_file(config_path)
    lpc = LinchpinContext()
    lpc.load_config(search_path=[config_path])
    lpc.load_global_evars()
    lpc.setup_logging()

    lpa = LinchpinAPI(lpc)


    filter = GenericInventory.GenericInventory(role_path=lpa.role_path)

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    res_output = 'linchpin.benchmark'
    res_file = open(mock_path+'/'+res_output)
    res_output = json.load(res_file)['17']['targets'][0]['general-inventory']['outputs']['resources']
    res_file.close()

def setup_generic_topology():
    global topology

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    topo = 'PinFile'
    topo_file = open(mock_path+'/'+topo)
    topology = yaml.load(topo_file, Loader=yaml.FullLoader)['general-inventory']['topology']
    topo_file.close()

def setup_generic_config():
    global config

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    cfg = 'PinFile'
    cfg_file = open(mock_path+'/'+cfg)
    config = yaml.load(cfg_file, Loader=yaml.FullLoader)['general-inventory']['cfgs']
    cfg_file.close()

def setup_generic_layout():
    global layout

    provider = 'general-inventory'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    template = 'linchpin.benchmark'
    template_file = open(mock_path+'/'+template)
    layout = json.load(template_file)['17']['targets'][0]['general-inventory']['inputs']['layout_data']['inventory_layout']

def setup_complex_workspace():
    global workspace

    provider = 'complex-inventory'
    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    workspace = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)


@with_setup(setup_generic_inventory_filter)
@with_setup(setup_generic_config)
def test_get_host_data():
    """
    """
    host_data = filter.get_host_data(res_output, config)
    hosts = []
    for data in host_data:
        for host in data:
            assert_true('__IP__' in list(data[host].keys()))
            hosts.append(data[host]['__IP__'])
    expected_hosts = ['10.179.254.83', 'hp-dl380pgen8-02-vm-15.lab.bos.redhat.com', 'dummy-744068-0', 'dummy-744068-1', 'dummy-744068-2', '192.168.122.219', '172.16.100.11', '155.132.55.17', '11.50.197.1', '109.254.93.117', '180.20.194.96', '194.231.24.112', '171.109.242.199', 'test.example.com', '192.168.122.210']
    assert_equal(set(hosts), set(expected_hosts))


@with_setup(setup_generic_inventory_filter)
@with_setup(setup_generic_config)
def test_get_hosts_by_count():
    """
    """
    host_data = filter.get_host_data(res_output, config)
    expected_hosts = ['10.179.254.83', 'hp-dl380pgen8-02-vm-15.lab.bos.redhat.com', 'dummy-744068-0']
    hosts = filter.get_hosts_by_count(host_data, 0)
    assert_equal(len(hosts), 0)
    hosts = filter.get_hosts_by_count(host_data, 3)
    assert_equal(len(hosts), 3)
    assert_equal(set(hosts), set(expected_hosts))


@with_setup(setup_generic_inventory_filter)
@with_setup(setup_generic_topology)
@with_setup(setup_generic_config)
@with_setup(setup_generic_layout)
def test_get_inventory():
    """
    """
    inventory = filter.get_inventory(res_output, layout, topology, config)
    assert_true(inventory)



#!/usr/bin/env python

# flake8: noqa

from __future__ import absolute_import
from nose.tools import assert_equal

import linchpin.InventoryFilters
from linchpin.InventoryFilters import InventoryProviders



from __future__ import absolute_import
from collections import OrderedDict

from linchpin.InventoryFilters.InventoryFilter import InventoryFilter


class Inventory(InventoryFilter):
    DEFAULT_HOSTNAMES = ['system']

    def get_host_data(self, res, cfgs):
        """
        Returns a dict of hostnames or IP addresses for use in an Ansible
        inventory file, based on available data. Only a single hostname or IP
        address will be returned per instance, so as to avoid duplicate runs of
        Ansible on the same host via the generated inventory file.

        Each hostname contains mappings of any variable that was defined in the
        cfgs section of the PinFile (e.g. __IP__) to the value in the field that
        corresponds with that variable in the cfgs.

        By default, the hostname will be the system field returned by beaker

        :param topo:
            linchpin Beaker resource data

        :param cfgs:
            map of config options from PinFile
        """

        host_data = OrderedDict()
        if res['resource_group'] != 'beaker' or res['role'] != 'bkr_server':
            return host_data
        var_data = cfgs.get('beaker', {})
        if var_data is None:
            var_data = {}
        host = self.get_hostname(res, var_data,
                                 self.DEFAULT_HOSTNAMES)
        hostname_var = host[0]
        hostname = host[1]
        host_data[hostname] = {}
        if '__IP__' not in list(var_data.keys()):
            var_data['__IP__'] = hostname_var
        self.set_config_values(host_data[hostname], res, var_data)
        return host_data



from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.plugins.action import ActionBase
import linchpin.MockUtils.MockUtils as mock_utils


class ActionModule(ActionBase):
    def run(self, tmp=None, task_vars=None):
        """
        Simple action plugin that returns the mocked output
        when linchpin_mock is True
        """
        super(ActionModule, self).run(tmp, task_vars)
        module_args = self._task.args.copy()
        linchpin_mock = task_vars['vars'].get('linchpin_mock',
                                              False)
        if linchpin_mock:
            return mock_utils.get_mock_data(module_args,
                                            "ec2_vpc_net_facts")

        module_return = self._execute_module(module_args=module_args,
                                             task_vars=task_vars, tmp=tmp)
        return module_return



#!/usr/bin/env python
from string import digits, ascii_letters


def memory_to_int(memory):
    amount = ''.join(c for c in memory if c in digits)
    amount = int(amount)
    size = ''.join(c for c in memory if c in ascii_letters)
    if size == 'Gi':
        amount = amount * 1000
    return amount


class FilterModule(object):
    ''' A filter to add_res_data '''
    def filters(self):
        return {
            'memory_to_int': memory_to_int
        }



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to provide default value when not given '''
    def filters(self):
        return {
            'provide_default': filter_utils.provide_default
        }



from linchpin import LinchpinAPI
from linchpin.context import LinchpinContext
from linchpin.exceptions import LinchpinError
import yaml
import json
import os
import tempfile


class Workspace(object):

    def __init__(self,
                 path=None):

        """
        Linchpin api workspace constructor

        :param path: path to workspace directory
        """

        self.workspace_path = path
        self.context = LinchpinContext()
        self.context.setup_logging()
        self.context.load_config(workspace=path)
        self.context.load_global_evars()
        self.pindict = {}
        # check where the workspace is being created
        self.context.set_cfg('lp', 'workspace', self.workspace_path)
        self.context.set_evar('workspace', self.workspace_path)
        self.context.set_evar('debug_mode', True)
        self.context.set_evar('auth_debug', True)

    def load_data(self, path):
        """
        load_data function to load from workspace path

        :param path: path to workspace directory
        """

        self.pindict = yaml.load(open(path).read())
        return self.pindict

    def validate(self):
        """
        validate function to validate loaded workspace/pinfile

        """

        linchpin_api = LinchpinAPI(self.context)
        self.load_data(self.find_pinfile())
        output = linchpin_api.do_validation(self.pindict)
        return output

    def find_pinfile(self):
        """
        find_pinfile function to search pinfiles in workspace path
        returns pinfile path if found

        """

        PF_NAMES = ["Pinfile", "PinFile", "PinFile.json"]
        for name in PF_NAMES:
            if os.path.isfile(self.workspace_path + "/" + name):
                return self.workspace_path + "/" + name
        return False

    def set_workspace(self, path):
        """
        set_workspace function sets workspace path

        :param path: path to workspace directory

        returns workspace path if set
        """

        self.workspace_path = path
        self.context.set_cfg('lp', 'workspace', self.workspace_path)
        self.context.set_evar('workspace', self.workspace_path)
        return self.workspace_path

    def get_workspace(self):
        """
        get_workspace function gets current workspace path

        :param path: path to workspace directory

        returns workspace path if set
        """


        return self.workspace_path

    def set_evar(self, key, value):
        """
        set_evar function sets extra vars in current run

        :param key: string
        :param value: string


        returns key,value tuple
        """

        self.context.set_evar(key, value)
        return key, value

    def get_evar(self, key):
        """
        get_evar function sets extra vars in current run

        :param key: string


        returns value for corresponding key
        """

        return self.context.get_evar(key)

    def set_credentials_path(self, creds_path):
        """
        set_credentials_path function set credentials path

        :param creds_path: path to credential directory

        returns True/False
        """

        if os.path.isdir(creds_path):
            return self.context.set_evar("default_credentials_path",
                                         creds_path)
        raise LinchpinError("Incorrect file path, path should be a directory")

    def get_credentials_path(self):
        """
        get_credentials_path function gets current credentials path

        returns path to credential file
        """

        return self.context.get_evar("default_credentials_path")

    def set_vault_encryption(self, vault_enc):
        """
        set_vault_encryption sets vault_encryption flag
        if credentials are encrypted in vault current credentials path

        param: vault_enc: boolean

        returns boolean
        """

        if isinstance(vault_enc, bool):
            return self.context.set_evar("vault_encryption", vault_enc)
        raise LinchpinError("Incorrect datatype please use boolean")

    def get_vault_encryption(self):
        """
        get_vault_encryption gets current vault_encryption flag value

        returns boolean
        """

        return self.context.get_evar("vault_encryption")

    def set_flag_no_hooks(self, flag):
        """
        set_flag_no_hooks sets no_hooks flag

        param: flag: boolean

        returns boolean
        """

        if isinstance(flag, bool):
            return self.context.set_cfg("hookflags", "no_hooks", flag)
        raise LinchpinError("Incorrect datatype please use boolean")

    def get_flag_no_hooks(self):
        """
        get_flag_no_hooks gets current vault_encryption flag value

        returns boolean
        """

        return self.context.get_cfg("hookflags", "no_hooks")

    def set_flag_ignore_failed_hooks(self, flag):
        """
        set_flag_ignore_failed_hooks sets current ignore_failed_hooks flag value

        param: flag: boolean

        """

        if isinstance(flag, bool):
            return self.context.set_cfg("hookflags",
                                        "ignore_failed_hooks",
                                        flag)
        raise LinchpinError("Incorrect datatype please use boolean")

    def get_flag_ignore_failed_hooks(self):
        """
        get_flag_ignore_failed_hooks get current ignore_failed_hooks flag value

        returns boolean

        """

        return self.context.get_cfg("hookflags", "ignore_failed_hooks")

    def set_vault_pass(self, vault_pass):
        """
        set_vault_pass set current vault_pass value

        param: vault_pass: string
        returns boolean

        """

        return self.context.set_evar("vault_password", vault_pass)

    def get_vault_pass(self):
        """
        get_valut_pass get current valut_password set

        returns boolean

        """

        return self.context.get_evar("vault_password")

    def get_inventory(self, inv_format="json"):
        """
        get_inventory gets inventory of latest run

        param: inv_format: string json/ini

        returns dict/string

        """

        lapi = LinchpinAPI(self.context)
        inventory_data = lapi._write_to_inventory(inv_format=inv_format)
        return inventory_data

    def get_latest_run(self):
        """
        get_latest_run get latest resources provisioned

        returns dict

        """

        lapi = LinchpinAPI(self.context)
        latest_run_data = lapi._get_run_data_by_txid()
        return latest_run_data

    def set_cfg(self, section, key, value):
        """
        get_flag_ignore_failed_hooks get current ignore_failed_hooks flag value

        returns boolean

        """

        return self.context.set_cfg(section, key, value)

    def get_cfg(self, section, key):
        """
        get_cfg gets current linchpin.conf values based on section, key

        returns string

        """

        return self.context.set_cfg(section, key)

    def up(self):
        """
        provisions workspace resources constructed through the workspace object

        returns output dictionary

        """

        lapi = LinchpinAPI(self.context)
        file_path = self.find_pinfile()
        self.load_data(file_path)
        output = lapi.do_action(self.pindict, action='up')
        return output

    def destroy(self):
        """
        Destroys workspace resources constructed through the workspace object

        returns output dictionary

        """

        lapi = LinchpinAPI(self.context)
        self.load_data(self.find_pinfile())
        output = lapi.do_action(self.pindict, action='destroy')
        return output


class Pinfile(Workspace):

    def __init__(self,
                 pinfile={},
                 config="linchpin.conf",
                 workspace_path=None):

        """
        Linchpin api Pinfile constructor

        :param pinfile: dictionary object of pinfile
        :param config: configuration path to linchpin.conf, defaults to
                       current working directory
        :param workspace_path: path to workspace directory. if not provided
                               workspace would be generated in /tmp/
        """

        # if workspace is not provided
        if workspace_path:
            self.workspace_path = workspace_path
        else:
            self.workspace_path = tempfile.mkdtemp()
        # generate a workspace name based on Pinfile_contents
        # write contents to workspace
        open(self.workspace_path + "/PinFile", "w").write(json.dumps(pinfile))
        self.pinfile = pinfile
        self.config = config
        self.context = LinchpinContext()
        self.context.setup_logging()
        self.context.load_config(workspace=self.workspace_path)
        self.context.load_global_evars()
        # check where the workspace is being created
        self.context.set_cfg('lp', 'workspace', self.workspace_path)
        self.context.set_evar('workspace', self.workspace_path)
        self.context.set_evar('debug_mode', True)
        self.context.set_evar('auth_debug', True)

    def up(self):
        """
        provsions pinfile resources constructed through the Pinfile object

        returns output dictionary
        """
        lapi = LinchpinAPI(self.context)
        output = lapi.do_action(self.pinfile, action='up')
        return output

    def destroy(self):
        """
        Destroys pinfile resources constructed through the Pinfile object

        returns output dictionary

        """

        lapi = LinchpinAPI(self.context)
        output = lapi.do_action(self.pinfile, action='destroy')
        return output



#!/usr/bin/env python

from __future__ import absolute_import
from collections import OrderedDict

from linchpin.InventoryFilters.InventoryFilter import InventoryFilter


class Inventory(InventoryFilter):
    DEFAULT_HOSTNAMES = ['public_dns_name', 'public_ip', 'private_ip']

    def get_host_data(self, res, cfgs):
        """
        Returns a dict of hostnames or IP addresses for use in an Ansible
        inventory file, based on available data. Only a single hostname or IP
        address will be returned per instance, so as to avoid duplicate runs of
        Ansible on the same host via the generated inventory file.
        Each hostname contains mappings of any variable that was defined in the
        cfgs section of the PinFile (e.g. __IP__) to the value in the field that
        corresponds with that variable in the cfgs.
        If an instance has a public IP attached, its hostname in DNS will be
        returned if available, and if not the public IP address will be used.
        For instances which have a private IP address for VPC use cases, the
        private IP address will be returned since private EC2 hostnames (e.g.
        ip-10-0-0-1.ec2.internal) will not typically be resolvable outside of
        AWS. For instances with both a public and private IP address, the
        public address is always returned instead of the private address.
        :param topo:
            linchpin AWS EC2 resource data
        :param cfgs:
            map of config options from PinFile
        """

        host_data = OrderedDict()
        if res['resource_group'] != 'aws' or res['role'] != 'aws_ec2':
            return host_data
        var_data = cfgs.get('aws', {})
        if var_data is None:
            var_data = {}
        for instance in res['instances']:
            host = self.get_hostname(instance, var_data,
                                     self.DEFAULT_HOSTNAMES)
            hostname_var = host[0]
            hostname = host[1]
            if '__IP__' not in list(var_data.keys()):
                var_data['__IP__'] = hostname_var
            host_data[hostname] = {}
            self.set_config_values(host_data[hostname], instance, var_data)
        return host_data



"""
Inventory Providers List
"""
from __future__ import absolute_import
from linchpin.exceptions import LinchpinError
from .JSONInventoryFormatter import JSONInventoryFormatter
from .CFGInventoryFormatter import CFGInventoryFormatter

formatter_classes = {
    "cfg": CFGInventoryFormatter,
    "ini": CFGInventoryFormatter,
    "json": JSONInventoryFormatter
}


def get_inv_formatter(inv_type):
    if inv_type not in formatter_classes:
        raise LinchpinError("Key {0} not found in"
                            " formatters".format(inv_type))
    return formatter_classes[inv_type]


def get_all_inv_formatters():
    return formatter_classes



from __future__ import absolute_import
from __future__ import print_function
import os
import json
from six.moves import range


def add_res_data(hosts, res_grp, role):
    new_hosts = []
    for host in hosts:
        host['resource_group'] = res_grp
        host['role'] = role
        new_hosts.append(host)
    return new_hosts


def ip_filter(forward_mode):
    if forward_mode == 'nat':
        return 'private'
    if forward_mode == 'bridge':
        return 'public'


def fetch_attr(output_dict, attr, default):
    return output_dict.get(attr, default)


def get_pod_status(pod_list, res_def_out):
    # this filter runs only when the kind is "Pod"
    pod_status = pod_list.get("resources", [])
    pod_stats = {}
    # filters out all the pods data w.r.to containers
    # provisioned in pods
    for x in pod_status:
        if x["metadata"]["name"] == res_def_out["result"]["metadata"]["name"]:
            pod_data = x
            pod_data["status"] = x["status"]["phase"]
            pod_stats[x["metadata"]["name"]] = pod_data
    # checks for the status of the pod and returns Failure value
    # when any pod is not in Running state
    for name in pod_stats.keys():
        if pod_stats[name]["status"] != "Running":
            return "Failure"
    return pod_stats


def omit_filter(output, omit):
    if output == "":
        return omit
    return output


def provide_default(fetched, default):
    if fetched == "":
        return default
    else:
        return fetched


def unicode_filter(output):
    output = json.dumps(output)
    return output


def format_rules(rules, rule_type):
    rules_output = []
    for rule in rules:
        if rule["rule_type"] == rule_type:
            rule_output = {}
            rule_output['from_port'] = rule['from_port']
            rule_output['to_port'] = rule['to_port']
            rule_output['cidr_ip'] = rule['cidr_ip']
            rule_output['proto'] = rule['proto']
            rules_output.append(rule_output)
    return rules_output


def fetch_list_by_attr(output, attr):
    new_output = []
    for ele in output:
        if attr in ele:
            new_output.append(ele[attr])
    return new_output


def get_host_from_uri(uri):
    """
    examples: qemu+ssh://root@hail.cloud.example.com/system
              test:///default
              qemu+ssh://192.168.122.6/system
    """
    uri = uri.split("//")[-1].split("/")[0].split("@")[-1]
    if uri == '':
        return 'localhost'
    return uri


def get_provider_resources(topo_output, res_type):
    provider_resources = []
    for host in topo_output:
        if host['resource_group'] == res_type:
            provider_resources.append(host)
    return provider_resources


def format_networks(networks):
    # "net-name=atomic-e2e-jenkins-test,net-name=atomic-e2e-jenkins-test2"
    nics = []
    if networks is not None and isinstance(networks, list):
        nics = ["net-name={0}".format(net) for net in networks]
        nics = ",".join(nics)

    return nics


def render_os_server_insts(res_def, res_def_names):

    output = []
    for sname in res_def_names:
        server_dict = {}
        server_dict['name'] = sname
        if 'additional_volumes' in res_def:
            server_dict['volumes'] = []
            for vol in res_def.get("additional_volumes", []):
                a_vol = {}
                a_vol.update(vol)
                a_vol["name"] = vol["name"] + "-" + server_dict["name"]
                a_vol["server_name"] = server_dict["name"]
                server_dict["volumes"].append(a_vol)
        output.append(server_dict)
    return output


def merge_two_dicts(x, y):
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z


def combine_hosts_names(hosts, names):
    result = []
    min_hosts_names = min(len(hosts), len(names))
    for i in range(min_hosts_names):
        result.append(merge_two_dicts(hosts[i], names[i]))
    if len(hosts) > min_hosts_names:
        for i in range(min_hosts_names, len(hosts)):
            result.append(hosts[i])
    if len(names) > min_hosts_names:
        for i in range(min_hosts_names, len(names)):
            result.append(names[i])
    return result


def filter_list_by_attr(output, attr):
    new_output = []
    for ele in output:
        if attr in ele:
            new_output.append(ele)
    return new_output


def get_libvirt_files(output):
    from xml.etree.ElementTree import fromstring
    files = []
    results = output['results']
    for result in results:
        if len(result['stdout']) > 0:
            stdout = result['stdout']
            myxml = fromstring(stdout)
            devices = myxml.findall('devices')
            for device in devices:
                disks = device.findall('disk')
                for disk in disks:
                    if disk.attrib["type"] == 'file':
                        if len(disk.findall('source')) > 0:
                            source = disk.findall('source')[0]
                            files.append(source.attrib['file'])
    return files


def translate_ruletype(ruletype):
    if ruletype == "inbound":
        return "ingress"
    if ruletype == "outbound":
        return "egress"
    else:
        return "invalid ruletype "


def filter_list_by_attr_val(output, attr, val):
    new_output = []
    for ele in output:
        if attr in ele:
            if ele[attr] == val:
                new_output.append(ele)
    return new_output


def get_network_domains(network, uri):
    import libvirt
    from xml.dom import minidom
    network_hosts = []
    conn = libvirt.open(uri)
    if conn is None:
        return network_hosts

    hosts = conn.listDomainsID()

    for host in hosts:
        dom = conn.lookupByID(host)
        raw_xml = dom.XMLDesc(0)
        xml = minidom.parseString(raw_xml)
        interfaces = xml.getElementsByTagName('interface')
        usesNetwork = False
        for interface in interfaces:
            usesNetwork = usesNetwork or iterate_interfaces(interface, network)
        if usesNetwork:
            network_hosts.append(dom.name())

    return network_hosts


def iterate_interfaces(interface, network):
    """
    Returns true if the interface uses the given network
    Otherwise returns false
    """
    if interface.getAttribute('type') != 'network':
        return False
    interfaceNodes = interface.childNodes
    for node in interfaceNodes:
        if node.nodeName != 'source':
            continue
        if 'network' not in list(node.attributes.keys()):
            return False
        if node.attributes['network'].nodeValue == network:
            return True
        return False


def map_results(results, attr, subattr):
    output = []
    for task_result in results:
        output.append(task_result[attr][subattr])
    return output


def prepare_ssh_args(ssh_args, users, sshkey):
    # "{{ ssh_args }} --ssh-inject \
    # {{ item.0}}:string:'{{ pubkey_local.stdout }}'"
    for user in users:
        ssh_args += "--ssh-inject " + user + ":string:'" + sshkey + "' "
    return ssh_args


def transform_os_server_output(res_def_out):
    res_def_out = res_def_out.get("results", [])
    res_def = {}
    res_def["ids"] = []
    res_def["openstack"] = []
    res_def["servers"] = []
    for ele in res_def_out:
        res_def["ids"].append(ele.get("id"))
        res_def["openstack"].append(ele.get("openstack", {}))
        res_def["servers"].append(ele.get("server", {}))
    return res_def


def fetch_beaker_job_ids(topo_out):
    output = []
    for entry in topo_out:
        entry_dict = {}
        entry_dict["ids"] = []
        # filter out if the resource's job id has already
        # been added. Workaround for Beaker server xmlrpc
        # issue on python2
        if "id" in list(entry) and not \
                [id for ed in output
                 for id in ed.get("ids", []) if entry.get("id") == id[2:]]:
            entry_dict["ids"].append("J:" + entry["id"])
            output.append(entry_dict)
    return output


def get_os_server_names(topo_output):
    names = []
    for item in topo_output:
        if item.get('role', None) == "os_server":
            openstack_res = item.get("openstack", [])
            for os_item in openstack_res:
                names.append(os_item["name"])
    return names


def write_to_file(data, path, filename, extn='.output', mode=0o600):
    filename = filename.replace(' ', '_').lower()
    path = os.path.expanduser(path)
    fd = open(path + filename + extn, "w")
    fd.write(json.dumps(data))
    fd.close()
    # change permissions to file
    # by default it is set to 0600
    os.chmod(path + filename + extn,
             mode)
    return data


def convert_to_json(input_str):
    if type(input_str) is str:
        return json.loads(input_str)
    else:
        return input_str


def path_relative_to(path, base_path):
    """
    If `path` is not an OS filesystem absolute or relative path then assume
    it is relative to `base_path`.
    """
    if path.startswith(('/', './', '../', '~/')):
        return path
    return os.path.join(base_path, path)



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as fil_utils


class FilterModule(object):
    ''' A filter fix distiller '''
    def filters(self):
        return {
            'transform_os_server_output': fil_utils.transform_os_server_output
        }



from __future__ import absolute_import
from __future__ import print_function
import os
import subprocess
import json
import tempfile
import shutil
from .action_manager import ActionManager
from cerberus import Validator

from linchpin.exceptions import HookError


class SubprocessActionManager(ActionManager):

    def __init__(self, name, action_data, target_data, state, **kwargs):

        """
        SubprocessActionManager constructor
        :param name: Name of Action Manager , ( ie., shell)
        :param action_data: dictionary of action_block
        consists of set of actions
        example:
        - name: hookname
          type: shell
          actions:
            - echo " this is post down operation Hello hai how r u ?"
            - test.sh
        :param target_data: Target specific data defined in PinFile
        :param kwargs: anyother keyword args passed as metadata
        """

        self.name = name
        self.action_data = action_data
        self.target_data = target_data
        self.context = kwargs.get('context', True)
        self.state = state
        self.kwargs = kwargs


    def validate(self):

        """
        Validates the action_block based on the cerberus schema
        example:
        action_block :: sample ::
        - name: manipulate_inventory
          type: shell
          path: /tmp/shellscripts
          actions:
            - thisisshell.sh
        """

        schema = {
            'name': {'type': 'string', 'required': True},
            'type': {
                'type': 'string',
                'allowed': ['shell', 'subprocess']
            },
            'path': {'type': 'string', 'required': False},
            'context': {'type': 'boolean', 'required': False},
            'vault_password_file': {'type': 'string', 'required': False},
            'src': {
                'type': 'dict',
                'schema': {
                    'type': {'type': 'string', 'required': True},
                    'url': {'type': 'string', 'required': True}
                }
            },
            'actions': {
                'type': 'list',
                'schema': {'type': 'string'},
                'required': True
            }
        }

        v = Validator(schema)
        status = v.validate(self.action_data)

        if not status:
            raise HookError("Invalid Syntax: {0}".format(str(v.errors)))
        else:
            return status


    def load(self):

        """
        Adds the shell script to the os path if mentioned
        """

        # set os.environpath if exists
        if 'path' in self.action_data:
            os.environ["PATH"] += ":{0}".format(self.action_data["path"])


    def add_context_params(self, action, results, data_path, context=True):

        """
        Adds ctx params to the action_block run when context is true
        :param file_path: path to the script
        :param context: whether the context params are to be included or not
        """

        command = action

        if self.context:
            data = ""
            for key in self.target_data:
                data += "{0}={1}; ".format(key, self.target_data[key])
            command = data + command
        return "{0} -- '{1}' {2}".format(command, results, data_path)


    def execute(self, results):

        """
        Executes the action_block in the PinFile
        """

        self.load()
        tmpdir = tempfile.mkdtemp()
        for action in self.action_data["actions"]:
            result = {}

            data_path = os.path.join(tmpdir, action)
            res_str = json.dumps(results, separators=(',', ':'))
            command = self.add_context_params(action,
                                              res_str,
                                              data_path,
                                              self.context)
            proc = subprocess.Popen(command,
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            proc.wait()

            for line in proc.stdout:
                print(line)

            try:
                data_file = open(data_path, 'r')
                data = data_file.read()
                if data:
                    result['data'] = json.loads(data)
            except IOError:
                # if an IOError is thrown, the file does not exist
                continue
            except ValueError:
                print("Warning: '{0}' is not a valid JSON object.  "
                      "Data from this hook will be discarded".format(data))
            result['return_code'] = proc.returncode
            result['state'] = str(self.state)
            results.append(result)
            data_file.close()

        shutil.rmtree(tmpdir)
        return results



from __future__ import absolute_import

from abc import ABCMeta, abstractmethod
import six


def usedb(func):
    def func_wrapper(*args, **kwargs):
        args[0]._opendb()
        x = func(*args, **kwargs)
        args[0]._closedb()
        return x
    return func_wrapper


class RunDB(six.with_metaclass(ABCMeta, object)):

    @abstractmethod
    def init_table(self, table):
        return self.driver.init_table(table)

    @abstractmethod
    def update_record(self, table, run_id, key, value):
        return self.driver.update_record(table, run_id, key, value)

    @abstractmethod
    def get_tx_record(self, table, tx_id):
        return self.driver.get_tx_record(table, tx_id)

    def get_tx_records(self, table, tx_ids):
        return self.driver.get_tx_records(table, tx_ids)

    @abstractmethod
    def get_run_id(self, table, action='up'):
        return self.driver.get_run_id(table, action=action)

    @abstractmethod
    def get_record(self, table, action='up', run_id=None):
        return self.driver.get_record(table, action=action, run_id=run_id)

    @abstractmethod
    def get_records(self, table=[], count=10):
        return self.driver.get_records(table=table, count=count)

    @abstractmethod
    def get_tables(self):
        return self.driver.get_tables()

    @abstractmethod
    def remove_record(self, table, key, value):
        return self.driver.remove_record(table, key, value)

    @abstractmethod
    def search(self, table, key):
        return self.driver.search(table, key)

    @abstractmethod
    def query(self, table, query_info):
        return self.driver.query(table, query_info)

    @abstractmethod
    def purge(self, table=None):
        return self.driver.purge(table)



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function
import os

from ansible.module_utils.basic import AnsibleModule

from bkr.common.hub import HubProxy
from bkr.common.pyconfig import PyConfigParser


BEAKER_CONF = os.environ.get('BEAKER_CONF', '/etc/beaker/client.conf')


class BeakerDistro(object):
    def __init__(self, logger=None):
        self.conf = PyConfigParser()
        default_config = os.path.expanduser(BEAKER_CONF)
        self.conf.load_from_file(default_config)
        self.hub = HubProxy(logger=logger, conf=self.conf)

    def search(self, params):
        return self.hub.distros.filter(params)


def main():

    argument_spec = dict(
        name=dict(type='str', required=False),
        arch=dict(type='str', required=False),
        family=dict(type='str', required=False),
        tags=dict(type='list', required=False),
        limit=dict(type='int', required=False, default=1),
    )
    mod = AnsibleModule(argument_spec)
    beaker = BeakerDistro()
    try:
        results = beaker.search(mod.params)
        mod.exit_json(hosts=results, changed=True)
    except Exception as ex:
        mod.fail_json(msg=str(ex), changed=True)


# import module snippets
main()



from __future__ import absolute_import
import os
import sys
import time
import yaml
import yamlordereddictloader
import shutil

from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_dict_equal
from nose.tools import assert_raises
from nose.tools import assert_dict_contains_subset
from nose.tools import with_setup

from linchpin import LinchpinAPI
from linchpin.cli import LinchpinCli
from linchpin.cli.context import LinchpinCliContext
from linchpin.tests.mockdata.contextdata import ContextData
from linchpin.exceptions import LinchpinError
from linchpin.exceptions import TopologyError

import six
from six.moves import zip


# ------------------------------- #
#    Setup functions for tests    #
# ------------------------------- #
def setup_load_config():

    """
    Perform setup of ContextData() object, and run get_temp_filename()
    """

    global cd
    global provider
    global config_path
    global config_data

    provider = 'dummy'

    cd = ContextData()
    cd.load_config_data()
    cd.create_config()
    config_path = cd.get_temp_filename()
    config_data = cd.cfg_data
    cd.write_config_file(config_path)


def setup_lp_cli():

    """
    Perform setup of LinchpinContext, lpc.load_config, and LinchPinAPI
    """

    global lpctx
    global lpc

    global target
    global pinfile

    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))

    setup_load_config()

    pinfile = 'PinFile'
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    lpctx = LinchpinCliContext()
    lpctx.load_config(lpconfig=config_path)
    lpctx.load_global_evars()
    lpctx.setup_logging()
    lpctx.workspace = os.path.realpath(mock_path)
    lpctx.pinfile = pinfile
    lpctx.pf_data = None
    lpctx.no_monitor = True
    lpctx.set_evar("no_monitor", True)
    lpctx.set_cfg("progress_bar", "no_progress", str(True))

    lpc = LinchpinCli(lpctx)
    lpc.disable_pbar = True
 


# ----------- #
#    Tests    #
# ----------- #
@with_setup(setup_lp_cli)
def test_get_data_path():
    # test the case in which the file is a full path that does not exist
    subdir = 'lp-test-get-data-path-{0}'.format(time.time())
    lpc.pf_data = '@/tmp/{0}'.format(subdir)
    assert_raises(TopologyError, lpc._get_data_path)



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to differentiate public and private ips '''
    def filters(self):
        return {
            'ip_filter': filter_utils.ip_filter
        }



# Without at least a file here, tests in the additional directory will not
# get picked up. If you add actual tests to this directory, then you can
# safely eliminate this file. Otherwise, it exists only to cause the tests in
# shared/tests to be discovered.
#
# Most tests should be written in the shared/tests directory so that they can
# be captured by all the scenarios. Only add tests here if there are tests
# only relevant to a particular scenario



import re
import glob, os
import requests
files=[]
texts= []
urls= []
print("We are scanning all the links! Please wait...")
os.chdir("./")
for file in glob.glob("*.rst"):
    files.append(open(file, "r+"))

for file in files:
    texts.append(file.read())

for text in texts:
    url= re.findall('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', text)
    urls=urls + url

brokenLinks=[]

count = 0
total = len(urls)
for url in urls:
    count+=1
    ret=0
    # url=url[0:len(url)-1]
    url = url.replace(">","")
    finish = float(count)/float(total)*100
    print("finished:"+" "+str(finish)[0:4]+"%")
    try:
        request = requests.get(url,timeout=3)
    except:
        ret = 999
    if ret != 999:
        ret = request.status_code
    if ret != 200:
        brokenLinks.append(url)

for link in brokenLinks:
    open("Deprecated_Link.txt","a").write(link+"\n")

print("Finished scanning! The deprecated links are in Deprecated_Link.txt")


#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to get provider resources '''
    def filters(self):
        return {
            'get_provider_resources': filter_utils.get_provider_resources
        }



#!/usr/bin/env python

import imp
import os

from .InventoryFilter import InventoryFilter
from .InventoryProviders import get_inv_formatter


class GenericInventory(InventoryFilter):

    def __init__(self, inv_format="cfg", role_path=None):
        InventoryFilter.__init__(self)
        self.role_path = role_path
        self.inv_formatter = get_inv_formatter(inv_format)()

    def get_host_data(self, res_output, config):
        """
        """
        host_data = []
        for res_grp in res_output:
            res_type = res_grp['resource_group']
            # check only when the inventory_filter exists
            filter_class = self.get_filter_class(res_type)
            data = filter_class.get_host_data(res_grp, config)
            host_data.append(data)
        return host_data


    def get_filter_class(self, res_type):
        path = "{0}/files/inventory.py".format(self._find_role_path(res_type))
        provider = imp.load_source('Inventory', path)
        return provider.Inventory()


    def _find_role_path(self, res_type):
        """
        returns the full path to the given playbook

        :params res_type: name of the role
        """

        for path in self.role_path:
            p = '{0}/{1}'.format(path, res_type)

            if os.path.exists(os.path.expanduser(p)):
                return p


    def get_hosts_by_count(self, host_data, count):
        """
        currently this function gets all the ips/hostname according to the
        order in which inventories are specified. later can be modified
        to work with user input
        """

        all_hosts = []
        for host in host_data:
            all_hosts.extend(list(host.keys()))
        return all_hosts[:count]

    def populate_config(self, host_dict, res_output, config):
        """
        """
        populated_config = {}
        for provider in host_dict.keys():
            for host in host_dict[provider].keys():
                populated_config[host] = {}
                for var in config[provider]:
                    value = self.filter_classes[provider]().\
                        get_field_values(res_output, var)
                    populated_config[host][var] = value
        return populated_config

    def get_inventory(self, res_output, layout, topology, config):
        # get all the topology host_ips
        host_data = self.get_host_data(res_output, config)
        # sort it based on topology
        # get the count of all layout hosts needed
        layout_host_count = self.get_layout_hosts(layout)
        # generate hosts list based on the layout host count
        inven_hosts = self.get_hosts_by_count(host_data,
                                              layout_host_count)
        # adding sections to respective host groups
        host_groups = self.get_layout_host_groups(layout)
        self.inv_formatter.add_sections(host_groups)
        # set children for each host group
        self.inv_formatter.set_children(layout)
        # set vars for each host group
        self.inv_formatter.set_vars(layout)
        # add ip addresses to each host
        self.inv_formatter.add_ips_to_groups(inven_hosts, layout)
        # add common vars to host_groups
        self.inv_formatter.add_common_vars(host_groups, layout, host_data)
        return self.inv_formatter.generate_inventory()



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to get host from libvirt uri '''
    def filters(self):
        return {
            'get_host_from_uri': filter_utils.get_host_from_uri
        }



#!/usr/bin/env python

# flake8: noqa

from __future__ import absolute_import
import os
import json
import yaml

from nose.tools import *
import linchpin.FilterUtils.FilterUtils as filter_utils


def test_add_res_data():
    hosts = [{}]
    new_hosts = filter_utils.add_res_data(hosts, "test_res_type", "test_role")
    assert_equal("test_res_type", new_hosts[0].get("resource_group"))
    assert_equal("test_role", new_hosts[0].get("role"))


def test_fetch_attr():
    test_dict = { "a": "b" }
    assert_equal("b", filter_utils.fetch_attr(test_dict, "a", "c"))
    assert_equal("d", filter_utils.fetch_attr(test_dict, "d", "d"))


def test_format_output():
    assert_equal("{}", filter_utils.unicode_filter({}))


def test_provide_default_omit():
    assert_equal("omit", filter_utils.provide_default("", "omit"))


def test_provide_default_val():
    assert_equal("a", filter_utils.provide_default("a", "omit"))


def test_get_pod_status():
    assert_equal({}, filter_utils.get_pod_status({}, {}))


def test_format_rules():
    rules = [{"rule_type": "inbound",
              "from_port": 22,
              "to_port": 22,
              "cidr_ip": "10",
              "proto": "tcp"
              }]
    expected = [{
              "from_port": 22,
              "to_port": 22,
              "cidr_ip": "10",
              "proto": "tcp"
               }]

    assert_equal(expected, filter_utils.format_rules(rules, "inbound"))


def test_fetch_list_by_attr():
    output = [{"a": "b"}, {"c","d"} ]
    assert_equal(["b"], filter_utils.fetch_list_by_attr(output, "a"))


def test_get_host_from_uri():
    test_str = "qemu+ssh://192.168.122.6/system"
    assert_equal("192.168.122.6", filter_utils.get_host_from_uri(test_str))


def test_get_host_from_uri_local():
    assert_equal("localhost", filter_utils.get_host_from_uri("test:///default"))


def test_get_provider_resources():
    test_out = [{"resource_group": "openstack"}]
    assert_equal(test_out,
                 filter_utils.get_provider_resources(test_out,
                                                     "openstack"))

def test_format_networks():
    networks = ["test"]
    assert_equal("net-name=test", filter_utils.format_networks(networks))


def test_os_server_insts():
    res_def = { 'additional_volumes': [{'name': 'test_vol', 'size': 1, 'device_name': '/dev/vdb'}]}
    expected = [{ "name": "test_vol-testname", 'size': 1, 'device_name': '/dev/vdb', 'server_name': 'testname'}]
    assert_equals(expected[-1],
                  filter_utils.render_os_server_insts(res_def, ["testname"])[-1]['volumes'][-1])


def test_merge_two_dicts():
    a = { "a": "b" }
    b = { "b": "c" }
    expected = {"a": "b", "b": "c"}
    assert_equals(expected, filter_utils.merge_two_dicts(a, b))


def test_combine_hosts_name():
    hosts = [{"a": "b"}, {"d": "e"}]
    names = [{"c": "d"}]
    expected = [{'a': 'b', 'c': 'd'}, {'d': 'e'}]
    assert_equals(expected, filter_utils.combine_hosts_names(hosts, names))


def test_filter_list_by_attr():
    output = [{"a": "b"}, {"d": "e"}]
    attr = "a"
    expected = [{"a":"b"}]
    assert_equals(expected, filter_utils.filter_list_by_attr(output, attr))


def test_translate_ruletype_inbound():
    assert_equals("ingress", filter_utils.translate_ruletype("inbound"))


def test_translate_ruletype_inbound():
    assert_equals("egress", filter_utils.translate_ruletype("outbound"))


def test_translate_ruletype_invalid():
    assert_equals("invalid ruletype ", filter_utils.translate_ruletype("x"))


def test_filter_list_by_attr_val():
    test_list = [{"a":"b"}, {"c":"d"}]
    expected = [{"a": "b"}]
    assert_equals(expected,
                  filter_utils.filter_list_by_attr_val(test_list, "a", "b"))


def test_map_results():
    test = [{ "a": {"b": "c"} }]
    assert_equals(["c"], filter_utils.map_results(test, "a", "b"))


def test_prepare_ssh_args():
    ssh_args = ""
    users = ["testuser"]
    sshkey = "xyz"
    assert_equals("--ssh-inject testuser:string:'xyz' ",
                  filter_utils.prepare_ssh_args("",users, sshkey))


def test_transform_os_server_output():
    res_def_out = {"results": [{ "id": "testid" }]}
    expected = {'ids': ['testid'], 'openstack': [{}], 'servers': [{}]}
    assert_equals(expected,
                  filter_utils.transform_os_server_output(res_def_out))

def test_fetch_beaker_job_ids():
    test_input = [{"id": "3124"}, {"id": "3214"}]
    expected = [{'ids': ['J:3124']}, {'ids': ['J:3214']}]
    assert_equals(expected,
                  filter_utils.fetch_beaker_job_ids(test_input))


def test_fetch_beaker_job_ids_count():
    test_input = [{"id": "3124"}, {"id": "3124"}, {"id": "3214"}]
    expected = [{'ids': ['J:3124']}, {'ids': ['J:3214']}]
    assert_equals(expected,
                  filter_utils.fetch_beaker_job_ids(test_input))

def test_fetch_os_server_names():
    test_input = [ {"resource_group": "openstack",
                    "role": "os_server",
                    "openstack": [{"name": "test"}]
                   }
                 ]
    assert_equals(["test"],
                  filter_utils.get_os_server_names(test_input))



from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.plugins.action import ActionBase
import linchpin.MockUtils.MockUtils as mock_utils
import zmq
import sys

if sys.version_info >= (3, 3):
    from unittest.mock import MagicMock
else:
    from mock import MagicMock


def instance_id(result):
    return result['instance_id']


class ActionModule(ActionBase):
    ''' Send events to Nanomsg '''

    TRANSFERS_FILES = False

    def run(self, tmp=None, task_vars=None):

        module_args = self._task.args.copy()
        if task_vars is None:
            task_vars = dict()
        linchpin_mock = task_vars['vars'].get('linchpin_mock',
                                              False)
        if linchpin_mock:
            return mock_utils.get_mock_data(module_args,
                                            "ec2")

        vm = self._task.get_variable_manager()
        if vm.extra_vars.get('no_monitor', False):
            def get_dict(context, key):
                return context.__context_dict[key]

            def set_dict(context, key, value):
                context.__context_dict[key] = value

            context = MagicMock()
            context.__context_dict = {}
            context.__getitem__ = get_dict
            context.__setitem__ = set_dict
        else:
            context = zmq.Context()
            socket = context.socket(zmq.REQ)
            socket.connect("tcp://localhost:5599")
            socket.RCVTIMEO = 2000

            if task_vars is None:
                task_vars = dict()
            if self._task.args.get('state', 'present') == 'present':
                action = 'present'
                base = self._task.args['instance_tags']['name']
                count = int(self._task.args.get('count', '1'))
                for index in range(1, count):
                    name = "{}({})".format(base, index)
                    msg = dict(status=dict(target=name, state='Provisioning'),
                               bar=10)
                    socket.send_string(str(msg))
                    socket.recv_string()
            elif self._task.args['state'] == 'absent':
                action = 'absent'
                name = "{}({})".format(self._task.args['instance_tags']['name'],
                                       self._task.args['instance_ids'])
                msg = dict(status=dict(target=name, state='Destryoing'), bar=10)
                socket.send(str(msg).encode('utf-8'))
                socket.recv_string()

            result = super(ActionModule, self).run(tmp, task_vars)
            result = self._execute_module(task_vars=task_vars,
                                          wrap_async=self._task.async_val)

            del tmp  # tmp no longer has any effect
            done = False
            old_state = ''
            while not done:
                facts_args = \
                    dict(instance_ids=result['instance_ids'],
                         aws_access_key=self._task.args['aws_access_key'],
                         aws_secret_key=self._task.args['aws_secret_key'],
                         region=self._task.args['region'])
                info = self._execute_module(
                    module_name='ec2_instance_facts',
                    module_args=facts_args,
                    task_vars=task_vars)
                done = True
                for instance in info['instances']:
                    state = instance['state']['name']
                    if state not in ['running', 'terminated']:
                        done = False
                    if state != old_state:
                        old_state = state
                        if state == 'pending':
                            bar = 40
                        if state == 'running':
                            state = 'Done'
                            bar = 50
                        if state == 'shutting-down':
                            bar = 40
                        if state == 'terminated':
                            state = 'Done'
                            bar = 50
                    else:
                        bar = 0
                    base = instance['tags']['name']
                    if action == 'present':
                        index = info['instances'].index(instance)
                    elif action == 'absent':
                        index = instance['instance_id']
                    name = "{}({})".format(base, index)
                    msg = dict(status=dict(target=name, state=state), bar=bar)
                    socket.send(str(msg).encode('utf-8'))
                    socket.recv_string()

            return result
        module_return = self._execute_module(module_args=module_args,
                                             task_vars=task_vars,
                                             tmp=tmp)

        return module_return



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to filter list by attribute '''
    def filters(self):
        return {
            'filter_list_by_attr': filter_utils.filter_list_by_attr
        }



#!/usr/bin/env python

from __future__ import absolute_import
import abc
import six
from six.moves import range

try:
    from configparser import ConfigParser
except ImportError:
    from six.moves.configparser import ConfigParser


class InventoryFilter(six.with_metaclass(abc.ABCMeta, object)):
    def __init__(self):
        self.config = ConfigParser(allow_no_value=True)

    @abc.abstractmethod
    def get_host_data(self, topo, config):
        pass

    def get_layout_hosts(self, inv):
        count = 0
        for host_group in inv['hosts']:
            count += host_group['count'] if 'count' in host_group else 1
        return count

    def get_layout_host_groups(self, inv):
        """
        get all the list of host groups in layout
        """
        host_groups = []
        for host in inv['hosts']:
            if "host_groups" in host:
                host_groups.extend(host["host_groups"])
        return list(set(host_groups))

    def add_sections(self, section_list):
        for section in section_list:
            self.config.add_section(section)
        # adding a default section all
        if "all" not in self.config.sections():
            self.config.add_section("all")

    def set_children(self, inv):
        if 'host_groups' not in list(inv.keys()):
            return
        for host_group in inv['host_groups']:
            if "children" in inv['host_groups'][host_group]:
                self.config.add_section("{0}:children".format(host_group))
                for child in inv['host_groups'][host_group]['children']:
                    self.config.set("{0}:children".format(host_group), child)


    def set_vars(self, inv):
        if 'host_groups' not in list(inv.keys()):
            return
        for host_group in inv['host_groups']:
            if "vars" in inv['host_groups'][host_group]:
                self.config.add_section("{0}:vars".format(host_group))
                for var in inv['host_groups'][host_group]['vars']:
                    grp_vars = inv['host_groups'][host_group]['vars'][var]
                    grp_vars = str(grp_vars)
                    self.config.set(host_group + ":" + "vars", var, grp_vars)
        for host in inv['hosts']:
            for host_group in host['host_groups']:
                self.config[host_group] = {}

    def add_ips_to_groups(self, inven_hosts, layout):
        inven_hosts.reverse()
        self.add_ips_to_host_group("all", inven_hosts)
        for host_name in layout['hosts']:
            if 'count' in list(host_name.keys()):
                count = host_name['count']
            else:
                count = 1
            host_list = []
            if count > len(inven_hosts):
                count = len(inven_hosts)
            for i in range(count, 0, -1):
                item = inven_hosts.pop()
                host_list.append(item)
            for host_group in host_name['host_groups']:
                self.add_ips_to_host_group(host_group, host_list)
        return True


    def add_ips_to_host_group(self, host_group, hosts):
        for ip in hosts:
            self.config.set(str(host_group), ip)


    def add_common_vars(self, host_groups, layout):
        # defaults common_vars to [] when they doesnot exist
        host_groups.append("all")
        common_vars = layout['vars'] if 'vars' in list(layout.keys()) else []
        for group in host_groups:
            items = list(dict(self.config.items(group)).keys())
            self.config.remove_section(group)
            self.config.add_section(group)
            for item in items:
                host_string = item
                for var in common_vars:
                    if common_vars[var] == "__IP__":
                        host_string += " " + var + "=" + item
                    else:
                        host_string += " " + var + "=" + common_vars[var]
                self.config.set(group, host_string)

    def get_hostname(self, data, cfgs, default_fields):
        if '__IP__' in list(cfgs.keys()):
            val = self.config_value_helper(data, cfgs['__IP__'])
            if val:
                return (cfgs['__IP__'], val)
        for var in default_fields:
            val = self.config_value_helper(data, var)
            if val:
                return (var, val)
        return ''

    def set_config_values(self, host_data, instance, cfgs={}):
        """
        finds the key associated with each config variable and get the
        value
        """
        if cfgs is None:
            return
        for var in cfgs.keys():
            host_data[var] = self.config_value_helper(instance, cfgs[var])

    def config_value_helper(self, instance, keys):
        """
        """
        if "." in keys:
            key, rest = keys.split('.', 1)
            # this handles errors in which the key does not exist
            if isinstance(instance, list) and key.isdigit():
                return self.config_value_helper(instance[int(key)], rest)
            if key not in list(instance.keys()):
                return ''
            return self.config_value_helper(instance[key], rest)
        else:
            if keys == '':
                return ''
            if keys not in list(instance.keys()):
                return ''
            return instance[keys]



from __future__ import absolute_import
from abc import ABCMeta, abstractmethod
import six

# action manager abstract class to be implemented by all
# action managers


class ActionManager(six.with_metaclass(ABCMeta, object)):

    @abstractmethod
    def validate(self):
        pass

    @abstractmethod
    def execute(self):
        pass



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to write variables to file '''
    def filters(self):
        return {
            'write_to_file': filter_utils.write_to_file
        }



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to format AWS EC2 security group rules '''
    def filters(self):
        return {
            'aws_sg_rules': filter_utils.format_rules
        }



from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible.plugins.action import ActionBase
import linchpin.MockUtils.MockUtils as mock_utils
from time import sleep
from mock import MagicMock
import zmq


class ActionModule(ActionBase):
    ''' Send events to Nanomsg '''

    TRANSFERS_FILES = False

    def run(self, tmp=None, task_vars=None):
        vm = self._task.get_variable_manager()
        if vm.extra_vars.get('no_monitor', False):
            def get_dict(context, key):
                return context.__context_dict[key]

            def set_dict(context, key, value):
                context.__context_dict[key] = value

            context = MagicMock()
            context.__context_dict = {}
            context.__getitem__ = get_dict
            context.__setitem__ = set_dict
            wait = 0
        else:
            context = zmq.Context()
            wait = 2
        socket = context.socket(zmq.REQ)
        socket.connect("tcp://localhost:5599")
        socket.RCVTIMEO = 2000

        if task_vars is None:
            task_vars = dict()
        name = "{}({})".format(self._task.args['name'],
                               self._task.args['count'])
        msg = dict(status=dict(target=name, state='Provisioning'), bar=50)
        socket.send_string(str(msg))
        socket.recv_string()

        super(ActionModule, self).run(tmp, task_vars)
        module_args = self._task.args.copy()
        linchpin_mock = task_vars['vars'].get('linchpin_mock',
                                              False)
        if linchpin_mock:
            return mock_utils.get_mock_data(module_args,
                                            "dummy")
        else:
            module_return = self._execute_module(module_name='dummy',
                                                 module_args=module_args,
                                                 task_vars=task_vars, tmp=tmp)
        del tmp  # tmp no longer has any effect
        sleep(wait)

        msg = dict(status=dict(target=name, state='Done'), bar=50)
        socket.send(str(msg).encode('utf-8'))
        socket.recv_string()
        sleep(wait)

        return module_return



#!/usr/bin/env python

from __future__ import absolute_import
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to get libvirt files from xml '''
    def filters(self):
        return {
            'get_libvirt_files': filter_utils.get_libvirt_files
        }



#!/usr/bin/env python

from __future__ import absolute_import
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to fix unicode format '''
    def filters(self):
        return {
            'unicode_filter': filter_utils.unicode_filter
        }



from __future__ import absolute_import
import os
import json
import shutil
import tempfile
from .action_manager import ActionManager
from Naked.toolshed.shell import muterun_js
from cerberus import Validator

from linchpin.exceptions import HookError


class NodejsActionManager(ActionManager):

    def __init__(self, name, action_data, target_data, state, **kwargs):

        """
        NodejsActionManager constructor
        :param name: Name of Action Manager , ( ie., nodejs)
        :param action_data: dictionary of action_block
        consists of set of actions
        example:
        - name: nameofhook
          type: nodejs
          context: true
          actions:
            - test.js
        :param target_data: Target specific data defined in PinFile
        :param kwargs: anyother keyword args passed as metadata
        """

        self.name = name
        self.action_data = action_data
        self.target_data = target_data
        self.context = kwargs.get('context', True)
        self.state = state
        self.kwargs = kwargs


    def validate(self):

        """
        Validates the action_block based on the cerberus schema
        """

        schema = {
            'name': {'type': 'string', 'required': True},
            'type': {'type': 'string', 'allowed': ['nodejs']},
            'path': {'type': 'string', 'required': False},
            'context': {'type': 'boolean', 'required': False},
            'vault_password_file': {'type': 'string', 'required': False},
            'src': {
                'type': 'dict',
                'schema': {
                    'type': {'type': 'string', 'required': True},
                    'url': {'type': 'string', 'required': True}
                }
            },
            'actions': {
                'type': 'list',
                'schema': {'type': 'string'},
                'required': True
            }
        }

        v = Validator(schema)
        status = v.validate(self.action_data)

        if not status:
            raise HookError("Invalid syntax: {0}".format(+str((v.errors))))
        else:
            return status


    def add_ctx_params(self, hook_path, results, data_path, context=True):

        """
        Adds ctx params to the action_block run when context is true
        :param hook_path: path to the script
        :param context: whether the context params are to be included or not
        """
        params = hook_path
        if context:
            for key in self.target_data:
                params += " {0}={1}".format(key, self.target_data[key])
        params += " -- '{0}' {1}".format(results, data_path)
        return params


    def execute(self, results):

        """
        Executes the action_block in the PinFile
        """

        tmpdir = tempfile.mkdtemp()
        for action in self.action_data["actions"]:
            result = {}

            data_path = os.path.join(tmpdir, action)
            context = self.action_data.get("context", True)
            path = self.action_data["path"]
            hook_path = "{0}/{1}".format(path, action)
            res_str = json.dumps(results, separators=(',', ':'))
            command = self.add_ctx_params(hook_path,
                                          res_str,
                                          data_path,
                                          context)
            run_data = muterun_js(command, arguments=res_str)

            print(run_data.stdout)
            print(run_data.stderr)

            try:
                data_file = open(data_path, 'r')
                data = data_file.read()
                if data:
                    result['data'] = json.loads(data)
                data_file.close()
            except IOError:
                # if an IOError is thrown, the file was not created because
                # the hook passed no data back
                result['data'] = ''
            except ValueError:
                print("Warning: '{0}' is not a valid JSON object.  "
                      "Data from this hook will be discarded".format(data))

            result['return_code'] = run_data.exitcode
            result['state'] = str(self.state)
            results.append(result)

        shutil.rmtree(tmpdir)
        return results



from __future__ import print_function
import os
import json
import sys

output_path = sys.argv[-1]
print(output_path)
try:
    workspace = os.listdir(".")
    print("output to stderr: {0}".format(json.dumps(workspace)))
    out = open(output_path, 'w')
    out.write(json.dumps(workspace))
    out.close()
except OSError as e:
    print(e)



#!/usr/bin/env python

import os


def path_relative_to(path, base_path):
    """
    If `path` is not an OS filesystem absolute or relative path then assume
    it is relative to `base_path`.
    """
    if path.startswith(('/', './', '../', '~/')):
        return path
    return os.path.join(base_path, path)


class FilterModule(object):
    def filters(self):
        return {'path_relative_to': path_relative_to}



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to fetch jobids from topology outputs '''
    def filters(self):
        return {
            'fetch_beaker_job_ids': filter_utils.fetch_beaker_job_ids
        }



#!/usr/bin/env python

from __future__ import print_function
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to fix network format '''
    def filters(self):
        return {
            'map_results': filter_utils.map_results
        }



#!/usr/bin/env python

from __future__ import absolute_import
from collections import OrderedDict

from linchpin.InventoryFilters.InventoryFilter import InventoryFilter


class Inventory(InventoryFilter):
    def get_host_data(self, res, cfgs={}):
        """
        Returns a dict of hostnames or IP addresses for use in an Ansible
        inventory file, based on available data. Only a single hostname or IP
        address will be returned per instance, so as to avoid duplicate runs of
        Ansible on the same host via the generated inventory file.

        :param topo:
            linchpin Dummy/nummy resource data

        :param cfgs:
            map of config options from PinFile
        """
        host_data = OrderedDict()
        var_data = cfgs.get('dummy', {})
        var_data.update(cfgs.get('nummy', {}))
        if var_data is None:
            var_data = {}
            var_data['__IP__'] = '__SELF__'
        for host in res.get('hosts', []):
            host_data[host] = {}
            self.set_config_values(host_data[host], host, res, var_data)
        return host_data


    def set_config_values(self, host_data, host, instance, cfgs={}):
        """
        """
        if cfgs is None:
            return
        if 'hostname' not in list(cfgs.keys()):
            cfgs['hostname'] = '__IP__'
        for var in cfgs.keys():
            if var == 'hostname' and cfgs[var] == '__IP__':
                host_data[cfgs[var]] = host
            else:
                host_data[var] = self.config_value_helper(instance, cfgs[var])



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to translate rule_type '''
    def filters(self):
        return {
            'os_sg_rule_type': filter_utils.translate_ruletype
        }



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to add_res_data '''
    def filters(self):
        return {
            'add_res_data': filter_utils.add_res_data
        }



#!/usr/bin/env python

from __future__ import absolute_import
from six.moves import range
try:
    from StrinIO import StringIO
except ImportError:
    from io import StringIO

import collections

from .InventoryFormatter import InventoryFormatter

try:
    from configparser import ConfigParser
except ImportError:
    from six.moves.configparser import ConfigParser

from ipaddress import ip_interface


class CFGInventoryFormatter(InventoryFormatter):

    def __init__(self):
        InventoryFormatter.__init__(self)
        self.config = ConfigParser(allow_no_value=True)
        self.config.optionxform = str

    def add_sections(self, section_list):
        for section in section_list:
            self.config.add_section(section)
        # adding a default section all
        if "all" not in self.config.sections():
            self.config.add_section("all")

    def is_ipv6(self, ipaddr):
        try:
            if ip_interface(ipaddr).ip.version == 6:
                return True
            return False
        except Exception as e:
            print(e)
            return False

    def set_children(self, inv):
        if 'host_groups' not in list(inv.keys()):
            return
        for host_group in inv['host_groups']:
            if "children" in inv['host_groups'][host_group]:
                self.config.add_section("{0}:children".format(host_group))
                for child in inv['host_groups'][host_group]['children']:
                    self.config.set("{0}:children".format(host_group), child)

    def set_vars(self, inv):
        if 'host_groups' not in list(inv.keys()):
            return
        for host_group in inv['host_groups']:
            if "vars" in inv['host_groups'][host_group]:
                self.config.add_section("{0}:vars".format(host_group))
                for var in inv['host_groups'][host_group]['vars']:
                    grp_vars = inv['host_groups'][host_group]['vars'][var]
                    grp_vars = str(grp_vars)
                    self.config.set(host_group + ":" + "vars", var, grp_vars)


    def add_ips_to_groups(self, inven_hosts, layout):
        # create a ip to host mapping based on count
        inven_hosts.reverse()
        self.add_ips_to_host_group("all", inven_hosts)
        for host_name in layout['hosts']:
            if 'count' in list(host_name.keys()):
                count = host_name['count']
            else:
                count = 1
            host_list = []
            if count > len(inven_hosts):
                count = len(inven_hosts)
            for i in range(count, 0, -1):
                item = inven_hosts.pop()
                host_list.append(item)
            for host_group in host_name['host_groups']:
                self.add_ips_to_host_group(host_group, host_list)
        return True


    def add_ips_to_host_group(self, host_group, hosts):
        for ip in hosts:
            self.config.set(host_group, ip)


    def add_common_vars(self, host_groups, layout, config):
        # defaults common_vars to [] when they doesnot exist
        host_groups.append("all")
        common_vars = layout['vars'] if 'vars' in list(layout.keys()) else []
        for group in host_groups:
            items = list(collections.OrderedDict(self.config.items(group))
                                    .keys())
            self.config.remove_section(group)
            self.config.add_section(group)
            for item in items:
                host_string = item
                for var in common_vars:
                    for cfg_item in config:
                        if item not in list(cfg_item.keys()):
                            continue
                        if common_vars[var] in list(cfg_item[item].keys()):
                            value = common_vars[var]
                            ip_addr = str(cfg_item[item][value])
                            if ((var == "ansible_ssh_host") or
                                    (var == "ansible_host"))\
                                    and self.is_ipv6(ip_addr):
                                host_string += " " + \
                                               var + \
                                               "=" + \
                                               "[" + \
                                               str(cfg_item[item][value]) + "]"
                            else:
                                host_string += " " + \
                                               var + \
                                               "=" + \
                                               str(cfg_item[item][value])
                        else:
                            if ((var == "ansible_ssh_host") or
                                    (var == "ansible_host")) and\
                                    self.is_ipv6(str(common_vars[var])):
                                host_string += " " + \
                                               var + \
                                               "=" + \
                                               "[" + str(common_vars[var]) + "]"
                            else:
                                host_string += " " + \
                                               var + \
                                               "=" + str(common_vars[var])
                self.config.set(group, host_string)

    def generate_inventory(self):
        output = StringIO()
        self.config.write(output)
        return output.getvalue()



from __future__ import absolute_import
from __future__ import print_function
from six.moves import range
import os
import json


def get_dummy_mock_data(module_args):
    hosts = [module_args['name'] + "-" + str(x)
             for x in range(0, int(module_args['count']))]
    data = {
        "changed": True,
        "dummy_file": "/tmp/dummy.hosts",
        "failed": False,
        "hosts": hosts,
        "output_type": "mock"
    }
    return data


def get_mock_data(module_args, module_name):
    dirname, filename = os.path.split(os.path.abspath(__file__))
    data = open(dirname + "/" + module_name + ".data", "r").read()
    data = json.loads(data)
    return data



from __future__ import (absolute_import, division, print_function)
from ansible.plugins.callback import CallbackBase
__metaclass__ = type
import zmq

DOCUMENTATION = '''
    callback: zmq
    type: notification
    requirements:
      - zmq
    short_description: push to massage bus play events
    version_added: 1.9
    description:
      - This plugin will use zmq to push play events.
'''


class CallbackModule(CallbackBase):
    """
    makes Ansible much more exciting.
    """
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'notification'
    CALLBACK_NAME = 'zmq'
    CALLBACK_NEEDS_WHITELIST = False

    def __init__(self):

        super(CallbackModule, self).__init__()
        self.socket = None

    def push(self, msg):
        try:
            self.socket.send(str(msg).encode('utf-8'))
        except Exception:
            # FIXME: any other error
            pass

    def v2_playbook_on_play_start(self, playbook):
        vm = playbook.get_variable_manager()
        extra_vars = vm.extra_vars
        no_monitor = extra_vars.get('no_monitor', '')
        if no_monitor:
            self.disabled = True
        else:
            context = zmq.Context()
            self.socket = context.socket(zmq.REQ)
            self.socket.connect("tcp://localhost:5599")

    def v2_playbook_on_task_start(self, task, is_conditional):
        pass

    def v2_runner_on_ok(self, result):
        pass

    def playbook_on_stats(self, name):
        self.push('{"playbook": "done"}')



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to fix network format '''
    def filters(self):
        return {
            'os_net': filter_utils.format_networks
        }



from __future__ import absolute_import
import os
import sys
import time
import yaml
import yamlordereddictloader
import shutil

from nose.tools import assert_equal
from nose.tools import assert_false
from nose.tools import assert_dict_equal
from nose.tools import assert_dict_contains_subset
from nose.tools import with_setup

from linchpin import LinchpinAPI
from linchpin.cli import LinchpinCli
from linchpin.cli.context import LinchpinCliContext
from linchpin.tests.mockdata.contextdata import ContextData
from linchpin.exceptions import LinchpinError
from linchpin.exceptions import TopologyError

import six
from six.moves import zip


# ------------------------------- #
#    Setup functions for tests    #
# ------------------------------- #
def setup_load_config():

    """
    Perform setup of ContextData() object, and run get_temp_filename()
    """

    global cd
    global provider
    global config_path
    global config_data

    provider = 'dummy'

    cd = ContextData()
    cd.load_config_data()
    cd.create_config()
    config_path = cd.get_temp_filename()
    config_data = cd.cfg_data
    cd.write_config_file(config_path)


def setup_lp_api():

    """
    Perform setup of LinchpinContext, lpc.load_config, and LinchPinAPI
    """

    global lpc
    global lpa

    global target
    global pinfile

    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))

    setup_load_config()

    pinfile = 'PinFile'
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    lpc = LinchpinCliContext()
    lpc.load_config(lpconfig=config_path)
    lpc.load_global_evars()
    lpc.setup_logging()
    lpc.workspace = os.path.realpath(mock_path)

    lpa = LinchpinAPI(lpc)


def setup_lp_cli():

    """
    Perform setup of LinchpinContext, lpc.load_config, and LinchPinAPI
    """

    global lpctx
    global lpc

    global target
    global pinfile

    base_path = '{0}'.format(os.path.dirname(
        os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))

    setup_load_config()

    pinfile = 'PinFile'
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)

    lpctx = LinchpinCliContext()
    lpctx.load_config(lpconfig=config_path)
    lpctx.load_global_evars()
    lpctx.setup_logging()
    lpctx.workspace = os.path.realpath(mock_path)
    lpctx.pinfile = pinfile
    lpctx.pf_data = None
    lpctx.no_monitor = True
    lpctx.set_evar("no_monitor", True)
    lpctx.set_cfg("progress_bar", "no_progress", str(True))

    lpc = LinchpinCli(lpctx)
    lpc.disable_pbar = True
 


# ----------- #
#    Tests    #
# ----------- #
def test_cli_create():

    lpctx = LinchpinCliContext()
    lpctx.load_config()
    lpctx.load_global_evars()
    lpc = LinchpinCli(lpctx)

    assert_equal(isinstance(lpc, LinchpinCli), True)


@with_setup(setup_lp_cli)
def test_pinfile():
    pf = {}
    pf['foo'] = 'bar'

    # set the pinfile
    lpc.pinfile = pf

    # get the pinfile and verify that the data is correct
    data = lpc.pinfile

    assert_dict_equal(pf, data)


@with_setup(setup_lp_cli)
def test_pf_data():
    pf_data = {}
    pf_data['foo'] = 'bar'

    # set the pinfile
    lpc.pf_data = pf_data

    # get the pinfile and verify that the data is correct
    data = lpc.pf_data

    assert_dict_equal(pf_data, data)


@with_setup(setup_lp_cli)
def test_workspace():
    workspace = '/tmp/workspace'

    # set the pinfile
    lpc.workspace = workspace

    # get the pinfile and verify that the data is correct
    ws = lpc.workspace

    assert_equal(workspace, ws)


@with_setup(setup_lp_api)
def test_set_cfg():

    test_dict = {'key': 'value', 'key2': 'value2'}

    for k, v in test_dict.items():
        lpa.set_cfg('test', k, v)

    assert_dict_equal(test_dict, lpa.ctx.cfgs.get('test'))


@with_setup(setup_lp_api)
def test_get_cfg_section():

    lpa.set_cfg('test', 'key', 'value')
    lpa_dict = lpa.get_cfg('test')

    assert_dict_equal(lpa_dict, lpa.ctx.cfgs.get('test'))


@with_setup(setup_lp_api)
def test_get_cfg_item():

    lpa.set_cfg('test', 'key', 'value')
    lpa_value = lpa.get_cfg('test', 'key')

    test_dict = lpa.ctx.cfgs.get('test')
    test_value = test_dict.get('key')

    assert_equal(lpa_value, test_value)


@with_setup(setup_lp_api)
def test_set_evar():

    test_evars = {'ekey': 'evalue'}

    for k, v in test_evars.items():
        lpa.set_evar(k, v)
    lpa_evars = lpa.get_cfg('evars')

    assert_dict_contains_subset(test_evars, lpa_evars)


@with_setup(setup_lp_api)
def test_get_all_evars():

    test_evars = {'ekey1': 'evalue1', 'ekey2': 'evalue2'}
    for k, v in test_evars.items():
        lpa.set_evar(k, v)

    lpa_evars = lpa.get_evar()

    assert_dict_contains_subset(test_evars, lpa_evars)


@with_setup(setup_lp_cli)
def test_render_template():
    # load data as ordereddict
    provider = 'layouts'
    base_path = '{0}'.format(os.path.dirname(
    os.path.realpath(__file__))).rstrip('/')
    lib_path = os.path.realpath(os.path.join(base_path, os.pardir))
    mock_path = '{0}/{1}/{2}'.format(lib_path, 'mockdata', provider)
    layout = 'render_order.yml'
    layout_file = open(mock_path+'/'+layout, 'r')
    layout_data = lpc.parser.parse_json_yaml(layout_file.read())
    layout_file.close()

    # test that template leaves things that do not need to be templated
    template = lpc._render_template(layout_data, '{}')
    diff = [i1 for i1, i2 in zip(six.iteritems(layout_data), six.iteritems(template)) if i1 != i2]
    assert_false(diff)

    
    layout = 'template-inventory.yml'
    layout_file = open(mock_path+'/'+layout, 'r')
    layout_data = lpc.parser.parse_json_yaml(layout_file.read())
    layout_file.close()
    data = '{"executor_name": "lp_exec"}'
    template = lpc._render_template(layout_data, data)
    # test that template properly renders variables
    expected_hosts = ['lp_exec-master', 'lp_exec-node']
    hosts = set(template['inventory_layout']['hosts'])
    assert_equal(set(expected_hosts), hosts)
    # test that template turns variables without values into empty strings
    routehost = template['inventory_layout']['host_groups']['OSEv3']['vars']\
                ['openshift_hosted_registry_routehost']
    assert_equal(routehost, 'registry.cloudapps.')

    # test that returning no data will return the template
    template = {"one": "{{ two }}"}
    parsed = lpc._render_template(layout_data, None)


@with_setup(setup_lp_api)
def test_get_evar_item():

    test_evars = {'ekey2': 'evalue2'}
    for k, v in test_evars.items():
        lpa.set_evar(k, v)

    assert_equal(lpa.get_evar(key='ekey2'), test_evars.get('ekey2'))


@with_setup(setup_lp_cli)
def test_get_pinfile_path():
    try:
        lpc._get_pinfile_path()
    except LinchpinError as e:
        assert(e.contains('Please check that it exists and try again'))
    
    lpc.pinfile = 'PinFile.test'
    assert(lpc._get_pinfile_path(exists=False).endswith('PinFile.test'))   

    lpc.workspace = '/tmp'
    assert_equal(lpc._get_pinfile_path(exists=False),
                 '/tmp/PinFile.test')  


@with_setup(setup_lp_cli)
def test_get_data_path():
    # test the case in which the data is a string that doesn't start with '@' or
    # is none
    lpc.pf_data = None
    data_path = lpc._get_data_path()
    assert_equal(data_path, None)
    lpc.pf_data = '{}'
    data_path = lpc._get_data_path()
    assert_equal(data_path, None)

    # test the case in which the file is a full path
    lpc.pf_data = '@/tmp'
    data_path = lpc._get_data_path()
    assert_equal('/tmp', data_path)

    # test the case in which the file is not a full path and the workspace
    # is prepended
    lpc.pf_data = '@linchpin'
    data_path = lpc._get_data_path()
    assert(data_path.endswith('linchpin'))


@with_setup(setup_lp_cli)
def test_execute_action():
    ret_data = lpc._execute_action('up')
    assert_equal(ret_data[0], 0)
    run_id = list(ret_data[1])[0]
    tx_id = list(ret_data[1][run_id]['summary_data']['dummy'])[0]
    #lpc.pf_data = 'linchpin/tests/mockdata/templates/template-inventory.yml'
    ret_code = lpc._execute_action('up', run_id=run_id)[0]
    assert_equal(ret_code, 0)

    # test with use_pinfile == False
    lpc.ctx.set_cfg("lp", "use_rundb_for_actions", 'True')
    ret_code = lpc._execute_action('up', tx_id=tx_id)[0]
    # if we can get provision data into the rundb correctly we can make sure
    # this is 0 and increase test coverage as well
    assert_equal(ret_code, 99)


@with_setup(setup_lp_cli)
def test_find_include():
    lpc.workspace = os.path.realpath('linchpin/tests/mockdata/dummy')
    path = lpc.find_include('dummy-cluster.yml')
    assert(path.endswith('mockdata/dummy/topologies/dummy-cluster.yml'))

    path = lpc.find_include('dummy-layout.yml', ftype='layout')
    assert(path.endswith('mockdata/dummy/layouts/dummy-layout.yml'))

    path = lpc.find_include('dummy-hooks.yml', ftype='hooks')
    assert(path.endswith('mockdata/dummy/hooks/dummy-hooks.yml'))


@with_setup(setup_lp_cli)
def test_make_layout_integers():
    pf_w_path = lpc._get_pinfile_path()
    pf = lpc.parser.process(pf_w_path, data=None)
    provision_data = lpc._build(pf, pf_data=None)
    layout = provision_data['dummy']['layout']
    data = lpc._make_layout_integers(layout)
    field = data['inventory_layout']['hosts']['example-node']['count']
    assert_equal(type(field), int)

    # make count a string and try again
    provision_data = lpc._build(pf, pf_data=None)
    layout = provision_data['dummy']['layout']
    strfield = layout['inventory_layout']['hosts']['example-node']['count']
    strfield = str(strfield)
    data = lpc._make_layout_integers(layout)
    field = data['inventory_layout']['hosts']['example-node']['count']
    assert_equal(type(field), int)


@with_setup(setup_lp_cli)
def test_build():
    pf_w_path = lpc._get_pinfile_path()
    pf = lpc.parser.process(pf_w_path, data=None)
    provision_data = lpc._build(pf)

    # get topology file and compare it to provision_data['topology']
    topology_file = "{0}/topologies/dummy-cluster.yml".format(lpc.workspace)
    with open(topology_file, 'r') as topology_yaml:
        topology = yaml.load(topology_yaml,
                             Loader=yamlordereddictloader.Loader)
        assert_dict_equal(topology, provision_data['dummy']['topology'])
    # get layout file and compare it to provision_data['layout']
    layout_file = "{0}/layouts/dummy-layout.yml".format(lpc.workspace)
    with open(layout_file, 'r') as layout_yaml:
        layout = yaml.load(layout_yaml, Loader=yamlordereddictloader.Loader)
        assert_dict_equal(layout, provision_data['dummy']['layout'])
    # get hooks file and compare it to provision_data['hooks']
    hooks_file = "{0}/hooks/dummy-hooks.yml".format(lpc.workspace)
    with open(hooks_file, 'r') as hooks_yaml:
        hooks = yaml.load(hooks_yaml, Loader=yamlordereddictloader.Loader)
        assert_dict_equal(hooks, provision_data['dummy']['hooks'])


@with_setup(setup_lp_cli)
def test_execute():
    # get provision data from PinFile
    pf_w_path = lpc._get_pinfile_path()
    pf = lpc.parser.process(pf_w_path, data=None)
    provision_data = lpc._build(pf, pf_data=None)
    # test with no targets
    ret_code = lpc._execute(provision_data, [])[0]
    assert_equal(ret_code, 0)

    # test with a specific target
    # rebuild because lpc._execute() or one of its sub-functions modifies
    # the provisioning data
    provision_data = lpc._build(pf, pf_data=None)
    ret_code = lpc._execute(provision_data, ['dummy'])[0]
    assert_equal(ret_code, 0)


@with_setup(setup_lp_cli)
def test_lp_fetch():
    # test where dest_ws == '.'
    # test where dest_ws != '.'
    # test where the path does not exist
    src = 'https://github.com/CentOS-PaaS-SIG/linchpin'
    root = 'linchpin/tests/mockdata/dummy'
    fetch_type = 'topologies'
    fetch_ref = 'develop'
    dest_ws = '/tmp/test-lp-fetch-{0}'.format(time.time())
    os.mkdir(dest_ws)

    # test where destination workspace != workdir
    
    lpc.lp_fetch(src,
                 root=root,
                 fetch_type=fetch_type,
                 fetch_ref=fetch_ref,
                 dest_ws=dest_ws)
    # check that the 'dummy' folder exists in dest_ws and that the only
    # sub-folder is 'topologies'
    cloned_ws = '{0}/dummy'.format(dest_ws)
    assert(os.path.exists(cloned_ws))
    children = os.listdir(cloned_ws)
    assert_equal(len(children), 1)
    assert_equal(children[0], 'topologies')
    shutil.rmtree(dest_ws)

    # test where destination workspace == None  
    os.mkdir(dest_ws)
    lpc.workspace = dest_ws
    lpc.lp_fetch(src,
                 root=root,
                 fetch_type='workspace',
                 fetch_protocol='FetchGit',
                 fetch_ref=fetch_ref,
                 dest_ws=None,
                 nocache=True)
    # check that the 'dummy' folder exists in cwd and that it contains the
    # sub-folders 'topologies' 'layouts' and 'hooks'
    assert(os.path.exists(cloned_ws))
    children = os.listdir(cloned_ws)
    assert('topologies' in children)
    assert('layouts' in children)
    assert('linchpin.conf' in children)
    assert('PinFile' in children)

    # clean up
    shutil.rmtree(dest_ws)


@with_setup(setup_lp_cli)
def test_lp_init():
    ws_path = '/tmp/lp-{0}'.format(time.time())
    lpc.pinfile = pinfile
    lpc.workspace = ws_path
    providers=['dummy']

    lpc.lp_init(providers=providers)


@with_setup(setup_lp_cli)
def test_lp_up_use_api():
    # disable hooks, which are causing the test to hang
    lpc.ctx.set_cfg("hook_flags", "no_hooks", True)

    run_data = lpc.lp_up()
    return_code = run_data[0]
    rundb_id = list(run_data[1])[0]
    assert_equal(return_code, 0)

    # TODO: make these lines work
    # If a run_id is provided, the run_id for the run should be the same so
    # that the run can be idempotent, but this doesn't work at the moment
    # run_data2 = lpc.lp_up(run_id=str(rundb_id))
    # assert_dict_equal(run_data[1], run_data2[1])


@with_setup(setup_lp_cli)
def test_lp_up_use_shell():
    # disable hooks, which are causing the test to hang
    lpc.ctx.set_cfg("hook_flags", "no_hooks", True)
    lpc.ctx.set_cfg("ansible", "use_shell", True)

    run_data = lpc.lp_up()
    return_code = run_data[0]
    rundb_id = list(run_data[1])[0]
    assert_equal(return_code, 0)

    # TODO: make these lines work
    # If a run_id is provided, the run_id for the run should be the same so
    # that the run can be idempotent, but this doesn't work at the moment
    # run_data2 = lpc.lp_up(run_id=str(rundb_id))
    # assert_dict_equal(run_data[1], run_data2[1])


@with_setup(setup_lp_cli)
def test_lp_validate():
    rc = lpc.lp_validate()[0]
    assert_equal(rc, 0)


@with_setup(setup_lp_cli)
def test_lp_setup():
    lpc.ctx.set_evar("use_venv", True)
    providers = ['docs', 'test']
    output = lpc.lp_setup(providers=providers)
    assert_equal(output[0], 0)


@with_setup(setup_lp_cli)
def test_lp_destroy():
    # disable hooks, which are causing the test to hang
    lpc.ctx.set_cfg("hook_flags", "no_hooks", True)
    lpc.lp_up()
    outputs = lpc.lp_destroy()
    return_code = outputs[0]
    assert_equal(return_code, 0)


@with_setup(setup_lp_cli)
def test_write_to_inventory():
    lpc._write_to_inventory()
    
    # test run with inventory path set
    inv_path = '/tmp/inventory-{0}'.format(time.time())
    lpc._write_to_inventory(inv_path=inv_path)


@with_setup(setup_lp_cli)
def test_write_distilled_context():
    # TODO: flesh this out
    lpc._write_distilled_context({})


def main():
    pass


if __name__ == '__main__':
    sys.exit(main())



from __future__ import absolute_import

from .tinydb import TinyRunDB

try:
    from .mongodb import MongoDB
except Exception as e:
    print("WARNING:" + str(e))
    print("WARNING: pymongo is currently not supported in python 2.7.x \
          Please Upgrade to Python 3.x If you would like to use Mongodb \
          Driver for rundb operations in linchpin")

try:
    DB_DRIVERS = {
        "TinyRunDB": TinyRunDB,
        "MongoDB": MongoDB,
    }
except Exception as e:
    DB_DRIVERS = {
        "TinyRunDB": TinyRunDB,
    }
    print("WARNING:" + str(e))
    print("WARNING: pymongo is currently not supported in python 2.7.x \
          Please Upgrade to Python 3.x If you would like to use Mongodb \
          Driver for rundb operations in linchpin")


def get_all_drivers():
    return DB_DRIVERS



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Author: Clint Savage- @herlo <herlo1@gmail.com>
#
# Manage the spin up / teardown of Duffy nodes within CentOS CI environment.
# See https://wiki.centos.org/QaWiki/CI/Duffy for more information about
# Duffy.
#

import json
import urllib
import os
from ansible.module_utils.basic import AnsibleModule

# --- Documentation Start ----------------------------------------------------#
DOCUMENTATION = '''
---
version_added: "0.1"
module: duffy
short_description: Duffy instance managemer
description:
  - This module allows a user to manage any number of duffy systems.
options:
  name:
    description:
      Given name for task
  state:
    description:
      Allocate or Deallocate instances
    required: true
  group:
    description:
      Group to which machine belongs. Useful when provisioning multiple machines
    default: name
  count:
    description:
      Number of instances to allocate
    default: 1
  key_path:
    description:
      Path to duffy.key
    default: ~/duffy.key
  version:
    description:
      Version of CentOS to provision
    choices: [5, 6, 7]
    default: 7
  arch:
    description:
    choices: [ 'x86_64', 'i386' ]
    default: 'x86_64'


notes:
  - See https://wiki.centos.org/QaWiki/CI/Duffy for more information
    required: true
requirements: []
author: Clint Savage - @herlo
'''

EXAMPLES = '''
- name: "provision 4 nodes"
  duffy:
    state: present
    count: 4
  register: herlo-ci

# teardown any nodes in ssid '171303a0'
- name: "teardown openshift nodes"
  duffy:
    state: absent
    ssid: {{ herlo-ci.ssid }}

'''

# ---- Logic Start ------------------------------------------------------------#


class Duffy:

    def __init__(self):
        pass

    def deallocate(self):
        res = urllib.urlopen(self.done_url).read()

        return {'status': res}

    def allocate(self):
        res = urllib.urlopen(self.get_url)

        try:
            b = json.load(res)
        except ValueError:
            raise Exception("The URL '{0}' \
                             produced an error.".format(self.get_url))

        # b={"hosts": ["n58.dusty.ci.centos.org",
        #               "n62.dusty.ci.centos.org"], "ssid": "e1c2d56e"}
        hosts = b['hosts']
        ssid = b['ssid']

        return hosts, ssid

    def execute(self, module):

        json_output = {}
        state = module.params['state']
        url_base = "http://admin.ci.centos.org:8080"
        api_key = open(os.path.expanduser(module.params['key_path'])).read()
        api_key = api_key.strip()

        # allocate some systems if state is 'present' :)
        if state == 'present':

            ver = module.params['version']
            arch = module.params['arch']
            count = module.params['count']

            self.get_url = "{0}/Node/get?key={1}&ver={2}&arch={3}\
                            &count={4}".format(url_base,
                                               api_key,
                                               ver,
                                               arch,
                                               count)
            (hosts, ssid) = self.allocate()
            json_output['hosts'] = hosts
            json_output['ssid'] = ssid

            return json_output
        elif state == 'absent':

            if not module.params.get('ssid'):
                module.fail_json(msg=str("The 'ssid' parameter is required."))

            ssid = module.params['ssid']

            self.done_url = "{0}/Node/done?key={1}&ssid={2}".format(url_base,
                                                                    api_key,
                                                                    ssid)
            return self.deallocate()


def main():

    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type='str'),
            state=dict(choices=['present', 'status', 'absent']),
            count=dict(default=1, type='int'),
            ssid=dict(default=None, type='str'),
            version=dict(default=7, type='int'),
            arch=dict(default='x86_64', type='str'),
            key_path=dict(default='~/duffy.key', type='str'),
        ),
    )

    try:
        d = Duffy()
        execute_output = d.execute(module)

        json_output = {}
        hosts = execute_output.get('hosts')
        status = execute_output.get('status')
        if hosts or status is not None:
            json_output['changed'] = True
            json_output.update(execute_output)
        else:
            json_output['changed'] = False

        module.exit_json(**json_output)
    except Exception as e:
        module.fail_json(msg=str(e))


# ---- Import Ansible Utilities (Ansible Framework) ---------------------------#
main()



#!/usr/bin/env python
from __future__ import absolute_import
from collections import OrderedDict

from linchpin.InventoryFilters.InventoryFilter import InventoryFilter


class Inventory(InventoryFilter):
    DEFAULT_HOSTNAMES = ["accessIPv4", "public_v4", "private_v4"]

    def get_host_data(self, res, cfgs):
        host_data = OrderedDict()
        if res['resource_group'] != 'openstack' or res['role'] != 'os_server':
            return host_data
        var_data = cfgs.get('openstack', {})
        if var_data is None:
            var_data = {}

        if 'results' in list(res.keys()):
            for result in res.get('results', []):
                if 'openstack' in list(result.keys()):
                    os_vars = result.get('openstack', [])
                    host = self.get_hostname(os_vars, var_data,
                                             self.DEFAULT_HOSTNAMES)
                    hostname_var = host[0]
                    hostname = host[1]
                    host_data[hostname] = {}
                    if '__IP__' not in list(var_data.keys()):
                        var_data['__IP__'] = hostname_var
                        host_data[hostname] = {}
                    self.set_config_values(host_data[hostname], os_vars,
                                           var_data)
                else:
                    continue
        else:
            grp = res.get('openstack', [])
            if isinstance(grp, list):
                for server in grp:
                    host = self.get_hostname(server, var_data,
                                             self.DEFAULT_HOSTNAMES)
                    hostname_var = host[0]
                    hostname = host[1]
                    host_data[hostname] = {}
                    if '__IP__' not in list(var_data.keys()):
                        var_data['__IP__'] = hostname_var
                        host_data[hostname] = {}
                    host_data[hostname] = {}
                    self.set_config_values(host_data[hostname], server,
                                           var_data)
            if isinstance(grp, dict):
                host = self.get_hostname(grp, var_data,
                                         self.DEFAULT_HOSTNAMES)
                hostname_var = host[0]
                hostname = host[1]
                host_data[hostname] = {}
                if '__IP__' not in list(var_data.keys()):
                    var_data['__IP__'] = hostname_var
                    host_data[hostname] = {}
                host_data[hostname] = {}
                self.set_config_values(host_data[hostname], grp, var_data)
        return host_data



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to get os_server_names from outputs '''
    def filters(self):
        return {
            'fetch_os_server_names': filter_utils.get_os_server_names
        }



#!/usr/bin/env python

from __future__ import absolute_import
from collections import OrderedDict
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from linchpin.InventoryFilters.InventoryFilter import InventoryFilter


class Inventory(InventoryFilter):
    DEFAULT_HOSTNAMES = ['ipv4', 'ipv6']

    def get_host_data(self, res, cfgs):
        """
        Returns a dict of hostnames or IP addresses for use in an Ansible
        inventory file, based on available data. Only a single hostname or IP
        address will be returned per instance, so as to avoid duplicate runs of
        Ansible on the same host via the generated inventory file.

        Each hostname contains mappings of any variable that was defined in the
        cfgs section of the PinFile (e.g. __IP__) to the value in the field that
        corresponds with that variable in the cfgs.

         :param topo:
            linchpin VMware guest resource data

        :param cfgs:
            map of config options from PinFile
        """

        host_data = OrderedDict()
        if res['resource_group'] != 'vmware' or res['role'] != 'vmware_guest':
            return host_data
        var_data = cfgs.get('vmware', {})
        if var_data is None:
            var_data = {}
        instance = res['instance']
        host = self.get_hostname(instance, var_data, self.DEFAULT_HOSTNAMES)
        hostname_var = host[0]
        hostname = host[1]
        if '__IP__' not in var_data.keys():
            var_data['__IP__'] = hostname_var
        host_data[hostname] = {}
        self.set_config_values(host_data[hostname], instance, var_data)
        return host_data

    def get_host_ips(self, host_data):
        if host_data:
            return host_data.keys()
        else:
            return []

    def get_inventory(self, topo, layout, config):
        host_data = []
        inven_hosts = []
        for res in topo:
            hd = self.get_host_data(res, config)
            if hd:
                host_data.append(hd)
            inven_hosts.extend(self.get_host_ips(hd))
        # adding sections to respective host groups
        host_groups = self.get_layout_host_groups(layout)
        self.add_sections(host_groups)
        # set children for each host group
        self.set_children(layout)
        # set vars for each host group
        self.set_vars(layout)
        # add ip addresses to each host
        self.add_ips_to_groups(inven_hosts, layout)
        self.add_common_vars(host_groups, layout)
        output = StringIO()
        self.config.write(output)
        return output.getvalue()



# The ansible_runner file contains all of the functions used to call ansible
# playbooks.  The runner can call playbooks via the Ansible API or the CLI,
# and supports the three most recent versions of Ansible (just as the Ansible
# team itself does)

from __future__ import absolute_import
import os
import sys
import json
import ansible
import subprocess
import tempfile
from contextlib import contextmanager

ansible24 = float(ansible.__version__[0:3]) >= 2.4
ansible_version = float(ansible.__version__[0:3])

# CentOS 6 EPEL provides an alternate Jinja2 package
# used by the imports below - Ansible uses Jinja2 here

try:
    from .callbacks import PlaybookCallback
    from ansible.parsing.dataloader import DataLoader
    from ansible.executor.playbook_executor import PlaybookExecutor
except ImportError:
    from .callbacks import PlaybookCallback
    from ansible.parsing.dataloader import DataLoader
    from ansible.executor.playbook_executor import PlaybookExecutor

if ansible24:
    from ansible.vars.manager import VariableManager
    from ansible.inventory.manager import InventoryManager as Inventory
else:
    from ansible.inventory import Inventory
    from ansible.vars import VariableManager

if ansible_version >= 2.8:
    from ansible import context


@contextmanager
def suppress_stdout():
    """
    This context manager provides tooling to make Ansible's Display class
    not output anything when used
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = devnull
        sys.stderr = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr


def subprocess_runner(cmd, shell=False):
    """
    Runs subprocess commands
    param: cmd in a list
    param: shell to print stdout, stderr or not

    """
    os.environ['PYTHONUNBUFFERED'] = "1"
    os.environ['ANSIBLE_FORCE_COLOR'] = '1'
    my_env = os.environ.copy()
    proc = subprocess.Popen(cmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True,
                            env=my_env
                            )
    stdout = []
    stderr = []
    while proc.poll() is None:
        std_out_line = proc.stdout.read()
        std_err_line = proc.stderr.read()
        if shell:
            print(std_out_line)
            print(std_err_line)
        else:
            stdout.append(std_out_line)
            stderr.append(std_err_line)
    return proc.returncode, stdout, stderr


def ansible_runner_24x(playbook_path,
                       extra_vars,
                       options,
                       inventory_src='localhost',
                       console=True):

    loader = DataLoader()
    variable_manager = VariableManager(loader=loader)
    variable_manager._extra_vars = extra_vars
    inventory = Inventory(loader=loader, sources=[inventory_src])
    variable_manager.set_inventory(inventory)
    passwords = {}

    pbex = PlaybookExecutor([playbook_path],
                            inventory,
                            variable_manager,
                            loader,
                            options,
                            passwords)
    return pbex


def ansible_runner_28x(playbook_path,
                       extra_vars,
                       options,
                       inventory_src='localhost',
                       console=True):

    loader = DataLoader()
    variable_manager = VariableManager(loader=loader)
    variable_manager._extra_vars = extra_vars
    inventory = Inventory(loader=loader, sources=[inventory_src])
    variable_manager.set_inventory(inventory)
    passwords = {}
    context._init_global_context(options)

    pbex = PlaybookExecutor([playbook_path],
                            inventory,
                            variable_manager,
                            loader,
                            passwords)
    return pbex


def set_environment_vars(env_vars):
    """
    Sets environment variables passed
    : param env_vars: list of tuples
    """
    for tup in env_vars:
        os.environ[tup[0]] = tup[1]
    return True


def ansible_runner_shell(playbook_path,
                         module_path,
                         extra_vars,
                         vault_password_file=None,
                         inventory_src='localhost',
                         verbosity=1,
                         console=True,
                         env_vars=(),
                         check=False):
    # set verbosity to >=2
    # since -v in subprocess command fails
    if verbosity >= 2:
        verbosity = verbosity
    else:
        verbosity = 2

    # set the base command
    base_command = ["ansible-playbook"]
    # convert verbosity to -v
    verbosity = "v" * verbosity
    verbosity = "-" + verbosity

    # append inventory option to command
    # FIX ME ansible considers localhost as another machine
    if os.path.isfile(inventory_src) and\
       inventory_src.split("/")[-1] != 'localhost':
        base_command.append("--inventory")
        base_command.append(inventory_src)

    # Load vault password file
    if vault_password_file and os.path.isfile(vault_password_file):
        base_command.append("--vault-password-file")
        base_command.append(vault_password_file)

    # ask for a sudo password
    if extra_vars.get('ask_sudo_pass', None):
        base_command.append("--ask-become-pass")

    # enable checkmode if check mode is mentioned
    if check:
        base_command.append("-C")


    # extravars to be passed into ansible.

    if extra_vars:

        fd, tmp_file_path = tempfile.mkstemp()
        with os.fdopen(fd, 'w') as tmp:
            # write files to temp file
            tmp.write(json.dumps(extra_vars))
        # Clean up the temporary file yourself
        base_command.append("-e")
        base_command.append("@" + tmp_file_path)

    base_command.append(verbosity)
    base_command.append(playbook_path)
    return_code, stdout, stderr = subprocess_runner(base_command,
                                                    shell=console)
    return return_code, stdout, stderr


def ansible_runner(playbook_path,
                   module_path,
                   extra_vars,
                   vault_password_file,
                   inventory_src='localhost',
                   verbosity=2,
                   console=True,
                   env_vars=(),
                   use_shell=False):
    """
    Uses the Ansible API code to invoke the specified linchpin playbook
    :param playbook: Which ansible playbook to run (default: 'up')
    :param console: Whether to display the ansible console (default: True)
    """

    # sets environment variables for subsequent processes
    if env_vars:
        set_environment_vars(env_vars)

    # note: It may be advantageous to put the options into the context and pass
    # that onto this method down the road. The verbosity flag would just live
    # in options and we could set the defaults.

    if not use_shell:

        connect_type = 'ssh'
        if 'localhost' in inventory_src:
            extra_vars["ansible_python_interpreter"] = sys.executable
            connect_type = 'local'

        options = Options(connect_type, module_path, 100, False, 'sudo', 'root',
                          False, False, False, False, None, None, None, None,
                          None, None, None, verbosity, False, False,
                          [vault_password_file])

        if ansible_version >= 2.8:
            pbex = ansible_runner_28x(playbook_path,
                                      extra_vars,
                                      options,
                                      inventory_src=inventory_src,
                                      console=console)
        elif ansible_version < 2.8:
            pbex = ansible_runner_24x(playbook_path,
                                      extra_vars,
                                      options,
                                      inventory_src=inventory_src,
                                      console=console)

        # ansible should be >=2.6 to run this statement
        cb = PlaybookCallback(options=options,
                              ansible_version=ansible_version)

        if not console:
            results = {}
            return_code = 0
            with suppress_stdout():
                pbex._tqm._stdout_callback = cb
            return_code = pbex.run()
            results = cb.results
            return return_code, results
        else:
            # the console only returns a return_code
            pbex._tqm._stdout_callback = None
            return_code = pbex.run()
            return return_code, None
    else:
        rc, stdout, stderr = ansible_runner_shell(playbook_path,
                                                  module_path,
                                                  extra_vars,
                                                  vault_password_file,
                                                  inventory_src=inventory_src,
                                                  verbosity=verbosity,
                                                  console=console)
        results = stdout
        return rc, results


# This is just a temporary class because python 2.7.5, the CentOS7 default,
# does not support the __dict__ keyword.  Once CentOS7 updates to a new version
# of python or we move to support CentOS8, we can delete this and move back to
# the (less messy) namedtuple or switch to the 3.7 DataClasses
class Options():
    def __init__(self, connection,
                 module_path,
                 forks, become, become_method, become_user, listhosts,
                 listtasks, listtags, syntax, remote_user, private_key_file,
                 ssh_common_args, ssh_extra_args, sftp_extra_args,
                 scp_extra_args, start_at_task, verbosity, check, diff,
                 vault_password_files):
        self.connection = connection
        self.module_path = module_path
        self.forks = forks
        self.become = become
        self.become_method = become_method
        self.become_user = become_user
        self.listhosts = listhosts
        self.listtasks = listtasks
        self.listtags = listtags
        self.syntax = syntax
        self.remote_user = remote_user
        self.private_key_file = private_key_file
        self.ssh_common_args = ssh_common_args
        self.ssh_extra_args = ssh_extra_args
        self.sftp_extra_args = sftp_extra_args
        self.scp_extra_args = scp_extra_args
        self.start_at_task = start_at_task
        self. verbosity = verbosity
        self.check = check
        self.diff = diff
        self.vault_password_files = vault_password_files



#!/usr/bin/env python


from __future__ import absolute_import
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to fix network format '''
    def filters(self):
        return {
            'render_os_server_insts': filter_utils.render_os_server_insts
        }



#!/usr/bin/env python
import linchpin.FilterUtils.FilterUtils as filter_utils


class FilterModule(object):
    ''' A filter to convert string to json '''
    def filters(self):
        return {
            'convert_to_json': filter_utils.convert_to_json
        }



GLOBAL_HOOKS = {
    "res_check": {
        "name": "res_check",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "res_check.yml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "ocp_container_boot_log": {
        "name": "ocp_container_boot_log",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "boot_log.yml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "ec2_boot_log": {
        "name": "ec2_boot_log",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "boot_log.yml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "azure_boot_log": {
        "name": "azure_boot_log",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "boot_log.yml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "beaker_log": {
        "name": "beaker_log",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "boot_log.yml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "os_server_boot_log": {
        "name": "os_server_boot_log",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "boot_log.yml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "check_ssh": {
        "name": "check_ssh",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "check_ssh.yaml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "ping": {
        "name": "ping",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "ping.yaml",
                "extra_vars": {
                    "ansible_ssh_user": "",
                    "ansible_ssh_private_key": ""
                }
            }
        ]
    },
    "port_up": {
        "name": "port_up",
        "context": True,
        "type": "ansible",
        "actions": [
            {
                "playbook": "port_up.yaml",
                "extra_vars": {
                    "port": 22,
                }
            }
        ]
    }
}



#!/usr/bin/env python

from __future__ import absolute_import
import os
import ast
import sys
import json
import time
import errno
import hashlib
import subprocess
import zmq

from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor

from uuid import getnode as get_mac
from collections import OrderedDict
from six import text_type

from linchpin.ansible_runner import ansible_runner
from linchpin import galaxy_runner

from linchpin.hooks.state import State
from linchpin.hooks import LinchpinHooks

from linchpin.rundb.drivers import DB_DRIVERS
from linchpin.rundb.basedb import BaseDB

from linchpin.exceptions import ActionError
from linchpin.exceptions import LinchpinError
from linchpin.exceptions import SchemaError
from linchpin.exceptions import ValidationError

from linchpin.validator import Validator

from linchpin.InventoryFilters import GenericInventory

if sys.version_info >= (3, 3):
    from unittest.mock import MagicMock
else:
    from mock import MagicMock


class LinchpinAPI(object):

    def __init__(self, ctx):
        """
        LinchpinAPI constructor

        :param ctx: context object from context.py

        """

        self.__meta__ = "API"

        self.ctx = ctx

        self.hook_state = None
        self._hook_observers = []
        self.playbook_pre_states = self.get_cfg('playbook_pre_states',
                                                {'up': 'preup',
                                                 'destroy': 'predestroy'})
        self.playbook_post_states = self.get_cfg('playbook_post_states',
                                                 {'up': 'postup',
                                                  'destroy': 'postdestroy',
                                                  'postinv': 'inventory'})
        self.hooks = LinchpinHooks(self)

        self.target_data = {}

        default_delimiter = self.get_cfg(section='evars',
                                         key='default_delimiter',
                                         default='_')

        base_path = '/'.join(os.path.dirname(__file__)
                                    .split('/')[0:-1])
        pkg = self.get_cfg(section='lp', key='pkg', default='linchpin')
        lp_path = '{0}/{1}'.format(base_path, pkg)
        self.pb_ext = self.get_cfg('extensions', 'playbooks', default='.yml')

        # get external_provider_path
        xp_path = self.get_cfg('lp',
                               'external_providers_path',
                               default='').split(':')


        pb_path = '{0}/{1}'.format(lp_path,
                                   self.get_evar('playbooks_folder',
                                                 default='provision'))

        role_path = '{0}/{1}/roles'.format(lp_path,
                                           self.get_evar('playbooks_folder',
                                                         default='provision'))
        self.role_path = [role_path]
        self.pb_path = pb_path

        for path in xp_path:
            role_path = '{0}/roles'.format(path)
            self.role_path.append(os.path.expanduser(role_path))

        for path in galaxy_runner.get_role_paths():
            self.role_path.append(os.path.expanduser(path))

        self.set_evar('default_delimiter', default_delimiter)
        self.set_evar('lp_path', lp_path)
        self.set_evar('pb_path', self.pb_path)
        self.set_evar('role_path', self.role_path)
        self.set_evar('from_api', True)
        self.set_evar('no_monitor', self.ctx.no_monitor)
        self.disable_pbar = 'True'
        self.workspace = self.get_evar('workspace')


    def setup_pbar(self):
        if (eval(self.get_cfg('progress_bar', 'no_progress')) or
           self.ctx.verbosity or self.ctx.no_monitor):
            self.disable_pbar = 'True'
        else:
            self.disable_pbar = 'False'
        self.pbar = tqdm_or_mock(disable=self.disable_pbar,
                                 bar_format="[{bar:10}] {desc:<30}|" +
                                            "{postfix[0][group]}",
                                 postfix=[dict(group='initializing')],
                                 position=0,
                                 leave=False)


    def setup_rundb(self):
        """
        Configures the run database parameters, sets them into extra_vars
        """

        rundb_conn = self.get_cfg(section='lp',
                                  key='rundb_conn')
        rundb_type = self.get_cfg(section='lp',
                                  key='rundb_type',
                                  default='TinyRunDB')
        rundb_conn_type = self.get_cfg(section='lp',
                                       key='rundb_conn_type',
                                       default='file')
        self.rundb_hash = self.get_cfg(section='lp',
                                       key='rundb_hash',
                                       default='sha256')

        if rundb_conn_type == 'file':

            rundb_conn_f = rundb_conn.replace('::mac::', str(get_mac()))
            rundb_conn_f = rundb_conn_f.replace('{{ workspace }}',
                                                self.workspace)
            rundb_conn_f = os.path.realpath(os.path.expanduser(rundb_conn_f))
            rundb_conn_dir = os.path.dirname(rundb_conn_f)

            if not os.path.exists(rundb_conn_dir):
                try:
                    os.makedirs(rundb_conn_dir)
                except OSError as exc:
                    if (exc.errno == errno.EEXIST and
                            os.path.isdir(rundb_conn_dir)):
                        pass
                    else:
                        raise
            rundb_conn = rundb_conn_f

        self.set_evar('rundb_type', rundb_type)
        self.set_evar('rundb_conn', rundb_conn)
        self.set_evar('rundb_hash', self.rundb_hash)

        return BaseDB(DB_DRIVERS[rundb_type], rundb_conn)


    def get_cfg(self, section=None, key=None, default=None):
        """
        Get cfgs value(s) by section and/or key, or the whole cfgs object

        :param section: section from ini-style config file

        :param key: key to get from config file, within section

        :param default: default value to return if nothing is found.
        """

        return self.ctx.get_cfg(section=section, key=key, default=default)


    def set_cfg(self, section, key, value):
        """
        Set a value in cfgs. Does not persist into a file,
        only during the current execution.

        :param section: section within ini-style config file

        :param key: key to use

        :param value: value to set into section within config file
        """

        self.ctx.set_cfg(section, key, value)


    def get_evar(self, key=None, default=None):
        """
        Get the current evars (extra_vars)

        :param key: key to use

        :param default: default value to return if nothing is found
        (default: None)
        """

        return self.ctx.get_evar(key=key, default=default)


    def set_evar(self, key, value):
        """
        Set a value into evars (extra_vars). Does not persist into a file,
        only during the current execution.

        :param key: key to use

        :param value: value to set into evars
        """

        self.ctx.set_evar(key, value)


    @property
    def hook_state(self):
        """
        getter function for hook_state property of the API object
        """

        return self.hook_state


    @hook_state.setter
    def hook_state(self, hook_state):
        """
        hook_state property setter , splits the hook_state string in
        subhook_state and sets linchpin.hook_state object

        :param hook_state: valid hook_state string mentioned in linchpin.conf
        """

        # call run_hooks after hook_state is being set
        if hook_state is None:
            return
        else:
            self.ctx.log_debug('hook {0} initiated'.format(hook_state))
            self._hook_state = State(hook_state, None, self.ctx)

            for callback in self._hook_observers:
                callback(self._hook_state)

    def bind_to_hook_state(self, callback):
        """
        Function used by LinchpinHooksclass to add callbacks

        :param callback: callback function
        """

        self._hook_observers.append(callback)


    def lp_journal(self, view='target', targets=[],
                   fields=None, count=1, tx_ids=None):

        rundb = self.setup_rundb()

        journal = {}

        if not len(targets):
            targets = rundb.get_tables()

        if view == 'target':
            # initialize rundb table
            dateformat = self.get_cfg('logger',
                                      'dateformat',
                                      default='%m/%d/%Y %I:%M:%S %p')
            for target in targets:
                tgt_records = rundb.get_records(table=target, count=count)

                if fields and 'start' in fields:
                    for run_id, record in tgt_records.items():
                        st = record.get('start')
                        strt = float(st) if len(st) else float(1000000000.0)
                        start = time.localtime(strt)
                        record['start'] = time.strftime(dateformat, start)

                if fields and 'end' in fields:
                    for run_id, record in tgt_records.items():
                        en = record.get('end')
                        endt = float(st) if len(en) else float(1000000000.0)
                        end = time.localtime(endt)
                        record['end'] = time.strftime(dateformat, end)


                journal[target] = tgt_records
        if view == 'tx':
            if len(tx_ids):
                journal = rundb.get_tx_records('linchpin', tx_ids)
            else:
                journal = rundb.get_records('linchpin', count=count)

        return journal


    def _get_role(self, role):
        for path in self.role_path:
            p = '{0}/{1}'.format(path, role)

            if os.path.exists(os.path.expanduser(p)):
                return

        # if the role is not in role_path, ansible-galaxy may be able to install
        # it (if it has not already).  galaxy_runner() will return True if the
        # role was successfully installed or if the role was previously
        # installed.  It will return false otherwise
        if not galaxy_runner.install(role):
            raise LinchpinError("role '{0}' not found in path: {1}\n. It also"
                                " could not be installed via Ansible"
                                " Galaxy".format(role, self.role_path))


    def _find_playbook_path(self, playbook):
        p = '{0}/{1}'.format(self.pb_path, playbook)

        if os.path.exists(os.path.expanduser(p)):
            return self.pb_path

        raise LinchpinError("playbook '{0}' not found in"
                            " path: {1}".format(playbook, self.pb_path))


    def _fix_broken_topologies(self, res_grp, res_grp_type):
        """
        Because the new tooling requires resource_definitions, both beaker
        and openshift didn't previously suport this section in topologies.
        This function is called if resource_definitions are not included
        in the topology. The topology will be given a resource_definitions
        section.

        :param res_grp: The resource group to update

        :param res_grp_type: Which type to convert (this may not matter)
        """

        res_defs = {}
        # with beaker, there will only be one
        # resource_definition upon conversion
        if res_grp_type == 'beaker':
            res_defs['role'] = 'bkr_server'
            res_defs['whiteboard'] = res_grp.pop('whiteboard',
                                                 'Provisioned with LinchPin')
            res_defs['job_group'] = res_grp.pop('job_group')
            res_defs['recipesets'] = res_grp.pop('recipesets')
            res_defs['cancel_message'] = res_grp.pop('cancel_message',
                                                     'Canceled by LinchPin')
            res_defs['max_attempts'] = res_grp.pop('max_attempts', 60)
            res_defs['attempt_wait_time'] = res_grp.pop('attempt_wait_time',
                                                        60)
            res_grp['resource_definitions'] = [res_defs]

            return res_grp

        if res_grp_type == 'openshift':
            res_defs = res_grp.pop('resources')

            count = 0
            for res in res_defs:
                res['name'] = 'openshift-res_{0}'.format(str(count))
                count += 1
                if res.get('inline_data'):
                    res['role'] = 'openshift_inline'
                    res['data'] = res.pop('inline_data')
                if res.get('file_reference'):
                    res['role'] = 'openshift_external'
                    res['filiename'] = res.pop('file_reference')

            res_grp['resource_definitions'] = res_defs

            # the old openshift role put credentials in the
            # resource group definition. Moving to credentials
            creds = {}
            creds['api_endpoint'] = res_grp.pop('api_endpoint')
            creds['api_token'] = res_grp.pop('api_token')
            res_grp['credentials'] = creds

            return res_grp


    def _convert_layout(self, layout_data):
        """
        Convert the layout to retain order of the layout hosts
        ;param layout_data: layout_data
        """
        layout_json = layout_data
        layout_hosts = []
        ihosts = layout_data["inventory_layout"]["hosts"]
        for k in ihosts:
            layout_host = {}
            layout_host["name"] = k
            for key in ihosts[k]:
                layout_host[key] = ihosts[k][key]
            layout_hosts.append(layout_host)
        layout_json["inventory_layout"]["hosts"] = layout_hosts
        return layout_json


    def generate_inventory(self, resource_data, layout, inv_format="cfg",
                           topology_data={}, config_data={}):
        inv = GenericInventory.GenericInventory(inv_format=inv_format,
                                                role_path=self.role_path)
        inventory = inv.get_inventory(resource_data, layout, topology_data,
                                      config_data)
        return inventory

    def get_pf_data_from_rundb(self, targets, run_id=None, tx_id=None):
        """
        This function takes the action and provision_data, returns the
        pinfile data

        :param targets: A list of targets for which to get the data

        :param targets: Tuple of target(s) for which to gather data.

        :param run_id: run_id associated with target (Default: None)

        :param tx_id: tx_id for which to gather data (Default: None)
        """

        rundb = self.setup_rundb()

        if run_id and tx_id:
            raise ActionError("'run_id' and 'tx_id' are mutually exclusive")

        pf_data = {}
        pinfile = OrderedDict()

        if run_id:
            for target in targets:
                pf_data[target] = rundb.get_record(target,
                                                   action='up',
                                                   run_id=run_id)
        if tx_id:
            record = rundb.get_tx_record('linchpin', tx_id)

            if not record or not len(record):
                return None

            if len(targets):
                for tgts in record['targets']:
                    for tgt, data in tgts.items():
                        run_id = int(list(data.keys())[0])
                        if tgt in targets:
                            tgt_data = (rundb.get_record(tgt,
                                        action=record['action'],
                                        run_id=run_id))
                            pf_data[tgt] = tgt_data
            else:
                for tgts in record['targets']:
                    for tgt, data in tgts.items():
                        run_id = int(list(data.keys())[0])
                        tgt_data = (rundb.get_record(tgt,
                                    action=record['action'],
                                    run_id=run_id))
                        pf_data[tgt] = tgt_data


        for t, data in pf_data.items():
            topo_data = data[0]['inputs'][0].get('topology_data')
            layout_data = data[0]['inputs'][0].get('layout_data')
            hooks_data = data[0]['inputs'][0].get('hooks_data')

            pinfile[t] = {}
            pinfile[t]['topology'] = topo_data
            pinfile[t]['run_id'] = data[1]
            if layout_data:
                pinfile[t]['layout'] = layout_data
            if hooks_data:
                pinfile[t]['hooks'] = hooks_data

        return pinfile


    def do_action(self, provision_data, action='up', run_id=None, tx_id=None):
        """
        This function takes provision_data, and executes the given
        action for each target within the provision_data disctionary.

        :param provision_data: PinFile data as a dictionary, with
        target information

        :param action: Action taken (up, destroy, etc). (Default: up)

        :param run_id: Provided run_id to duplicate/destroy (Default: None)

        :param tx_id: Provided tx_id to duplicate/destroy (Default: None)

        .. .note:: The `run_id` value differs from the `rundb_id`, in that
                   the `run_id` is an existing value in the database.
                   The `rundb_id` value is created to store the new record.
                   If the `run_id` is passed, it is used to collect an existing
                   `uhash` value from the given `run_id`, which is in turn used
                   to perform an idempotent reprovision, or destroy provisioned
                   resources.
        """

        results = {}
        return_code = 99

        # verify_targets_exist()
        targets = [x.lower() for x in list(provision_data.keys())]

        if 'linchpin' in targets:
            raise LinchpinError("Target 'linchpin' is not allowed.")

        for target in targets:
            if target == 'cfgs':
                continue
            if not isinstance(provision_data[target], dict):
                raise LinchpinError("Target '{0}' does not"
                                    " exist.".format(target))

            self.ctx.log_debug("Processing target: {0}".format(target))
            results[target] = {}
            self.set_evar('target', target)

            rundb_id = self.prepare_rundb(target, action, run_id, tx_id)

            try:
                validator = Validator(self.ctx, self.role_path, self.pb_ext)
                resources = validator.validate(provision_data[target])
            except SchemaError:
                raise ValidationError("Target '{0}' does not validate.  For"
                                      " more information run `linchpin "
                                      "validate`".format(target))

            topology_data = provision_data[target].get('topology')
            self.set_evar('topo_data', topology_data)

            self.update_rundb(rundb_id, target, provision_data)

            # note : changing the state triggers the hooks
            return_code, results[target]['task_results'] = (self.run_target(
                                                            target,
                                                            resources,
                                                            action,
                                                            run_id)
                                                           )

            end = time.time()

            self.rundb.update_record(target, rundb_id, 'end', str(end))
            self.rundb.update_record(target, rundb_id, 'rc', return_code)

            run_data = self.rundb.get_record(target,
                                             action=action,
                                             run_id=rundb_id)

            results[target]['rundb_data'] = {rundb_id: run_data[0]}

        # generate the linchpin_id and structure
        lp_data = self.write_results_to_rundb(results, action)

        return (return_code, lp_data)


    def run_target(self, target, resources, action, run_id=None):
        disable_uhash_targets = self.ctx.get_evar('disable_uhash_targets')
        enable_uhash = self.ctx.get_evar('enable_uhash')
        ansible_console = False
        if self.ctx.cfgs.get('ansible'):
            ansible_console = (
                ast.literal_eval(self.get_cfg('ansible',
                                              'console',
                                              default='False')))

        if not ansible_console:
            ansible_console = bool(self.ctx.verbosity)

        rundb_id = self.get_evar('rundb_id')
        if action == 'destroy':
            prev_id = run_id if run_id else self.rundb.get_run_id(target, 'up')
            self.hooks.rundb = (self.rundb, rundb_id, prev_id)
        else:
            self.hooks.rundb = (self.rundb, rundb_id)

        self.run_hooks('pre', action)

        # Enable/disable uhash based on provided flag(s)
        if disable_uhash_targets and target in disable_uhash_targets:
            self.ctx.set_evar('use_uhash', False)
        else:
            self.ctx.set_evar('use_uhash', enable_uhash)

        # invoke the appropriate action
        return_code, results = self._invoke_playbooks(resources,
                                                      action=action,
                                                      console=ansible_console)

        # logs the return_code back to console and quits
        # quits if return_code > 0
        self.ctx.log("Playbook Return code {0}".format(str(return_code)))
        if return_code == 0:
            self.ctx.log_state("Action '{0}' on Target '{1}' is "
                               "complete".format(action, target))
        elif action == "up":
            sys.exit(return_code)

        # FIXME Check the result[target] value here, and fail if applicable.
        # It's possible that a flag might allow more targets to run, then
        # return an error code at the end.

        self.run_hooks('inv', action)
        if self.__meta__ == "API":
            self.run_hooks('post', action)

        return return_code, results


    def run_hooks(self, state, action):
        self.pb_hooks = self.get_cfg('hookstates', action)
        if self.get_cfg('hook_flags', 'no_hooks'):
            return

        if state == 'inv':
            hook = 'postinv'
        else:
            hook = '{0}{1}'.format(state, action)
        self.ctx.log_debug('calling: {0}'.format(hook))
        if state in self.pb_hooks:
            self.hook_state = hook


    def write_results_to_rundb(self, results, action):
        lp_schema = ('{"action": "", "targets": []}')

        rundb = self.setup_rundb()
        rundb.schema = json.loads(lp_schema)
        lp_id = rundb.init_table('linchpin')

        summary = {}

        for target, data in results.items():
            for k, v in data['rundb_data'].items():
                str_k = str(k)
                summary[target] = {str_k: {'rc': v['rc'], 'uhash': v['uhash']}}

        rundb.update_record('linchpin', lp_id, 'action', action)
        rundb.update_record('linchpin', lp_id, 'targets', [summary])

        lp_data = {lp_id: {'action': action,
                           'summary_data': summary,
                           'results_data': results}}

        return lp_data


    def prepare_rundb(self, target, action, run_id=None, tx_id=None):
        self.rundb = self.setup_rundb()

        if tx_id:
            record = self.rundb.get_tx_record('linchpin', tx_id)
            run_id = (list(record['targets'][0][target].keys())[0])

        rundb_schema = json.loads(self.get_cfg(section='lp',
                                  key='rundb_schema'))
        self.rundb.schema = rundb_schema
        self.set_evar('rundb_schema', rundb_schema)

        start = time.time()
        st_uhash = int(start * 1000)
        uhash = None

        # generate a new rundb_id
        # (don't confuse it with an already existing run_id)
        rundb_id = self.rundb.init_table(target)
        orig_run_id = rundb_id

        uhash_length = self.get_cfg('lp', 'rundb_uhash_length')
        uhash_len = int(uhash_length)
        if not run_id:
            hash_str = ':'.join([target, str(tx_id), str(rundb_id),
                                 str(st_uhash)])
            if isinstance(hash_str, text_type):
                hash_str = hash_str.encode('utf-8')
            uh = hashlib.new(self.rundb_hash, hash_str)
            uhash = uh.hexdigest()[:uhash_len]

        if action == 'destroy' or run_id:
            # look for the action='up' records to destroy
            data, orig_run_id = self.rundb.get_record(target,
                                                      action='up',
                                                      run_id=run_id)

            if data:
                uhash = data.get('uhash')
                self.ctx.log_debug("using data from"
                                   " run_id: {0}".format(run_id))

        elif action not in ['up', 'destroy']:
            # it doesn't appear this code will will execute,
            # but if it does...
            raise LinchpinError("Attempting '{0}' action on"
                                " target: '{1}' failed. Not an"
                                " action.".format(action, target))


        self.ctx.log_debug('rundb_id: {0}'.format(rundb_id))
        self.ctx.log_debug('uhash: {0}'.format(uhash))

        self.rundb.update_record(target, rundb_id, 'uhash', uhash)
        self.rundb.update_record(target, rundb_id, 'start', str(start))
        self.rundb.update_record(target, rundb_id, 'action', action)

        self.set_evar('orig_run_id', orig_run_id)
        self.set_evar('rundb_id', rundb_id)
        self.set_evar('uhash', uhash)

        return rundb_id


    def update_rundb(self, rundb_id, target, provision_data):
        self.rundb.update_record(target,
                                 rundb_id,
                                 'inputs',
                                 [{'topology_data':
                                   provision_data[target]['topology']}
                                 ])

        if provision_data[target].get('layout', None):
            l_data = provision_data[target]['layout']
            provision_data[target]['layout'] = self._convert_layout(l_data)
            self.set_evar('layout_data', provision_data[target]['layout'])
            self.rundb.update_record(target,
                                     rundb_id,
                                     'inputs',
                                     [{'layout_data':
                                       provision_data[target]['layout']}
                                     ])

        if provision_data[target].get('hooks', None):
            hooks_data = provision_data[target].get('hooks')
            self.set_evar('hooks_data', hooks_data)
            self.rundb.update_record(target,
                                     rundb_id,
                                     'inputs',
                                     [{'hooks_data':
                                       provision_data[target]['hooks']}
                                     ])

        if provision_data.get('cfgs', None):
            vars_data = provision_data[target].get('cfgs')
            self.set_evar('cfgs_data', vars_data)
            self.rundb.update_record(target,
                                     rundb_id,
                                     'cfgs',
                                     [{'user':
                                      provision_data['cfgs']}
                                     ])


    def do_validation(self, provision_data, old_schema=False):
        """
        This function takes provision_data, and attempts to validate the
        topologies for that data

        :param provision_data: PinFile data as a dictionary, with
        target information
        """

        results = {}


        return_code = 0

        for target in list(provision_data.keys()):
            if not isinstance(provision_data[target], dict):
                raise LinchpinError("Target '{0}' does not"
                                    " exist.".format(target))

        targets = [x.lower() for x in list(provision_data.keys())]
        if 'linchpin' in targets:
            raise LinchpinError("Target 'linchpin' is not allowed.")

        for target in list(provision_data.keys()):
            if target == 'cfgs':
                continue

            self.ctx.log_debug("Processing target: {0}".format(target))

            results[target] = {}
            self.set_evar('target', target)

            validator = Validator(self.ctx, self.role_path, self.pb_ext)
            target_dict = provision_data[target]
            ret, results[target] = validator.validate_pretty(target_dict,
                                                             target,
                                                             old_schema)
            return_code += ret

        return return_code, results



    def _write_to_inventory(self, tx_id=None, inv_path=None, inv_format="cfg"):
        latest_run_data = self._get_run_data_by_txid()
        if tx_id:
            latest_run_data = self._get_run_data_by_txid(tx_id)
        all_inventories = {}
        try:
            targets = latest_run_data["targets"][0]
            for name in targets:
                if "layout_data" in targets[name]["inputs"]:
                    lt_data = targets[name]["inputs"]["layout_data"]
                    t_data = targets[name]["inputs"]["topology_data"]
                    c_data = {}
                    if "cfgs" in list(targets[name].keys()):
                        c_data = targets[name]["cfgs"]["user"]
                    layout = lt_data["inventory_layout"]
                    r_o = targets[name]["outputs"]["resources"]
                    """
                    # TODO: in the future we should render templates in
                    # layout and cfgs here so that we can use data from the
                    # most recent run
                    """
                    inv = self.generate_inventory(r_o,
                                                  layout,
                                                  inv_format=inv_format,
                                                  topology_data=t_data,
                                                  config_data=c_data)
                    all_inventories[name] = inv
            return all_inventories

        except Exception as e:
            self.ctx.log_state('Failed to write invenotry: {0}'.format(e))
            sys.exit(1)
        return True


    def get_run_data(self, tx_id, fields, targets=()):
        """
        Returns the RunDB for data from a specified field given a tx_id.
        The fields consist of the major sections in the RunDB (target
        view only). Those fields are action, start, end, inputs, outputs,
        uhash, and rc.

        :param tx_id: tx_id to search
        :param fields: Tuple of fields to retrieve for each record requested.
        :param targets: Tuple of targets to search from within the tx_ids
        """

        rundb = self.setup_rundb()

        tgt_run_ids = {}
        target_data = {}

        record = rundb.get_tx_record('linchpin', tx_id)

        if not record or not len(record):
            return None


        # get run_ids to query
        if len(targets):
            for tgts in record['targets']:
                for tgt, data in tgts.items():
                    if tgt in targets:
                        tgt_run_ids[tgt] = int(list(data.keys())[0])
        else:
            for tgts in record['targets']:
                for tgt, data in tgts.items():
                    tgt_run_ids[tgt] = int(list(data.keys())[0])

        for target, run_id in tgt_run_ids.items():
            record = rundb.get_record(target, run_id=run_id, action='up')
            field_data = {}
            single_value_fields = ('action',
                                   'start',
                                   'end',
                                   'rc',
                                   'uhash')

            for field in fields:
                f = record[0].get(field)
                if f:
                    if field in single_value_fields:
                        field_data[field] = f
                    else:
                        data_array = {}
                        for fld in f:
                            for k, v in fld.items():
                                if field == 'outputs':
                                    if isinstance(v, dict):
                                        values = v
                                    else:
                                        values = []
                                        for value in v:
                                            values.append(value)
                                    data_array[k] = values
                                else:
                                    data_array[k] = v

                            field_data[field] = data_array

            target_data[target] = field_data

        return target_data

    def _get_run_data_by_txid(self, tx_id=None):
        rundb = self.setup_rundb()
        latest_run_data = {}
        run_data = {}
        if tx_id is None:
            latest_run_data, tx_id = rundb.get_record('linchpin')
            run_data = self.get_run_data(tx_id,
                                         ('outputs',
                                          'inputs',
                                          'cfgs'))
        else:
            latest_run_data = rundb.get_record('linchpin',
                                               run_id=tx_id)
            latest_run_data = {tx_id: latest_run_data}
            run_data = self.get_run_data(tx_id,
                                         ('outputs',
                                          'inputs',
                                          'cfgs'))
        # this is expecting a dict of:
        # { tx_id: latest_run_data[tx_id] } but we're just giving it:
        # latest_run_data[tx_id]
        target_group = latest_run_data.get("targets", [])
        # Note:
        # target_group always returns a dict inside a list due to
        # rundb implemenatation. This code works for one PinFile run
        # might be subjected to change in future releases
        if len(target_group) == 0:
            return {}
        else:
            target_group = target_group[0]
        for key in list(target_group.keys()):
            target_group[key].update(run_data.get(key, {}))
        latest_run_data["targets"] = [target_group]
        return latest_run_data

    def _get_module_path(self):
        module_paths = []
        module_folder = self.get_cfg('lp',
                                     'module_folder',
                                     default='library')
        for path in reversed(self.role_path):
            module_paths.append('{0}/{1}/'.format(path, module_folder))
        return module_paths

    def _find_n_run_pb(self, pb_name, inv_src, console=True):
        self._get_role(pb_name)
        use_shell = ast.literal_eval(str(self.ctx.get_cfg("ansible",
                                                          "use_shell")))
        if pb_name == 'setup':
            pb_path = self._find_playbook_path("setup.yml")
            playbook_path = '{0}/{1}{2}'.format(pb_path, 'setup', self.pb_ext)
        else:
            pb_path = self._find_playbook_path("linchpin.yml")
            playbook_path = '{0}/{1}{2}'.format(pb_path,
                                                'linchpin',
                                                self.pb_ext)
        extra_vars = self.get_evar()
        env_vars = self.ctx.get_env_vars()
        if self.ctx.no_monitor:
            return_code, res = ansible_runner(playbook_path,
                                              self._get_module_path(),
                                              extra_vars,
                                              vault_password_file=None,
                                              inventory_src=inv_src,
                                              verbosity=self.ctx.verbosity,
                                              console=console,
                                              env_vars=env_vars,
                                              use_shell=use_shell
                                             )
        else:
            res = extra_vars['resources']
            group = res['resource_group_name']
            self.pbar.postfix[0] = dict(group="resource group '%s'" % group)
            self.pbar.refresh()
            with ProcessPoolExecutor() as executor:
                executor.submit(progress_monitor, self.disable_pbar, res)
                ansible_thread = executor.submit(
                    ansible_runner,
                    playbook_path,
                    self._get_module_path(),
                    extra_vars,
                    vault_password_file=None,
                    inventory_src=inv_src,
                    verbosity=self.ctx.verbosity,
                    console=console,
                    env_vars=env_vars,
                    use_shell=use_shell)

                return_code, res = ansible_thread.result()
            self.pbar.postfix[0] = dict(group="Finishing resource group " +
                                        group)
            self.pbar.refresh()
            self.pbar.update()
        return return_code, res

    def _invoke_playbooks(self, resources={}, action='up', console=True,
                          providers=[]):
        """
        Uses the Ansible API code to invoke the specified linchpin playbook

        :param resources: dict of resources to provision
        :param action: Which ansible action to run (default: 'up')
        :param console: Whether to display the ansible console (default: True)
        """

        return_code = 0
        results = []

        self.set_evar('_action', action)
        self.set_evar('state', 'present')

        if action == 'setup':
            self.set_evar('setup_providers', providers)
            return_code, res = self._find_n_run_pb(action,
                                                   "localhost",
                                                   console=console)
            if res:
                results.append(res)
            if not len(results):
                results = None
            return (return_code, results)

        if action == 'destroy':
            self.set_evar('state', 'absent')

        inventory_src = '{0}/localhost'.format(self.workspace)
        self.setup_pbar()
        self.pbar.set_description_str("Linchpin(%s)" % action)
        self.pbar.total = len(resources) + 1
        self.pbar.update()
        for resource in resources:
            self.set_evar('resources', resource)
            playbook = resource.get('resource_group_type')
            return_code, res = self._find_n_run_pb(playbook,
                                                   inventory_src,
                                                   console=console)
            if action == "up" and (return_code > 0 and
                                   not isinstance(return_code, str)):
                if self.ctx.verbosity > 0:
                    raise LinchpinError("Unsuccessful provision of resource")
                else:
                    res_grp_name = resource['resource_group_name']
                    print(res['failed'])
                    msg = res['failed'][0]._result['msg']
                    task = res['failed'][0].task_name
                    raise LinchpinError("Unable to provision resource group "
                                        "'{0}' due to '{1}' at task "
                                        " '{2}'".format(res_grp_name,
                                                        msg, task))
                sys.exit(return_code)

            if res:
                results.append(res)

        if not len(results):
            results = None

        self.pbar.postfix[0] = dict(group='Done')
        return (return_code, results)

    def ssh(self, target):
        def merge_two_dicts(x, y):
            z = x.copy()
            z.update(y)
            return z
        try:
            host = merge_two_dicts(
                {
                    'ansible_ssh_common_args': '',
                    'ansible_user': 'root',
                },
                self.ctx.inventory.hosts.get(target).serialize()['vars'])
        except AttributeError:
            if self.ctx.inventory.hosts.get(target) is None:
                print(target + ' was not found in the inventory')
                print('Found these hosts:')
                print('\n'.join(self.ctx.inventory.hosts.keys()))
                return

        if 'ansible_host' not in host:
            host['ansible_host'] = host['hostname']
        cmd = ('ssh {ansible_ssh_common_args}'
               '{ansible_user}@{ansible_host}')
        cmd = cmd.format(**host)
        print('connecting to {}: {}'.format(target, cmd))
        subprocess.call(cmd, shell=True)


def progress_monitor(disable_pbar, target):
    position = 1
    num_resources = 0
    resource_bar_format = "[{bar:10}] |-> {desc:<25}| {postfix[0][state]}"
    for resource in target['resource_definitions']:
        num_resources += resource.get('count', 1)
    group_bar = tqdm_or_mock(
        disable=disable_pbar,
        desc=target['resource_group_name'],
        bar_format="[{bar:10}] |-> {desc:<26}| {postfix[0][resource]}",
        postfix=[dict(resource='initializing')],
        total=num_resources,
        position=position, leave=False)
    resource_bars = dict()
    context = zmq.Context()
    socket = context.socket(zmq.REP)
    socket.bind("tcp://*:5599")

    event = dict(playbook='started')
    while event.get('playbook', 'in progress') != 'done':
        try:
            msg = socket.recv()
            msg = msg.decode("utf-8")
            event = ast.literal_eval(msg)
            resource = event['status']['target']
            group_bar.postfix[0] = dict(resource="resource '%s'" % resource)
            if resource not in resource_bars.keys():
                position += 1
                resource_bars[resource] = tqdm_or_mock(
                    disable=disable_pbar,
                    desc=resource,
                    bar_format=resource_bar_format,
                    postfix=[dict(state=event['status']['state'])],
                    total=100,
                    position=position, leave=False)
            resource_bars[resource].postfix[0] = event['status']
            resource_bars[resource].update(event['bar'])
            resource_bars[resource].refresh()
            if event['status']['state'] == 'Done':
                group_bar.update()
            socket.send_string("next", flags=1)
        except ValueError:
            continue
        except KeyError:
            continue
    group_bar.postfix[0] = dict(resource='Done')
    socket.close()
    context.destroy()
    return True


def tqdm_or_mock(disable, *args, **kwargs):
    if not eval(disable):
        return tqdm(*args, **kwargs)
    else:
        def get_dict(dbar, key):
            return dbar.__dbar_dict[key]

        def set_dict(dbar, key, value):
            dbar.__dbar_dict[key] = value

        dbar = MagicMock()
        dbar.__dbar_dict = {}
        dbar.__getitem__ = get_dict
        dbar.__setitem__ = set_dict
        return dbar
